from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias

from hugr.serialization import tys

from guppylang.error import InternalGuppyError
from guppylang.tys.common import ToHugr, Transformable, Transformer, Visitor
from guppylang.tys.const import BoundConstVar, Const, ConstValue, ExistentialConstVar
from guppylang.tys.var import ExistentialVar

if TYPE_CHECKING:
    from guppylang.tys.ty import Type


# We define the `Argument` type as a union of all `ArgumentBase` subclasses defined
# below. This models an algebraic data type and enables exhaustiveness checking in
# pattern matches etc.
# Note that this might become obsolete in case the `@sealed` decorator is added:
#  * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#  * https://github.com/johnthagen/sealed-typing-pep
Argument: TypeAlias = "TypeArg | ConstArg"


@dataclass(frozen=True)
class ArgumentBase(ToHugr[tys.TypeArg], Transformable["Argument"], ABC):
    """Abstract base class for arguments of parametrized types.

    For example, in the type `array[int, 42]` we have two arguments `int` and `42`.
    """

    @property
    @abstractmethod
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this argument."""


@dataclass(frozen=True)
class TypeArg(ArgumentBase):
    """Argument that can be instantiated for a `TypeParameter`."""

    # The type to instantiate
    ty: "Type"

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this argument."""
        return self.ty.unsolved_vars

    def to_hugr(self) -> tys.TypeArg:
        """Computes the Hugr representation of the argument."""
        return tys.TypeArg(tys.TypeTypeArg(ty=self.ty.to_hugr()))

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this argument."""
        if not visitor.visit(self):
            self.ty.visit(visitor)

    def transform(self, transformer: Transformer) -> Argument:
        """Accepts a transformer on this argument."""
        return transformer.transform(self) or TypeArg(self.ty.transform(transformer))


@dataclass(frozen=True)
class ConstArg(ArgumentBase):
    """Argument that can be substituted for a `ConstParam`."""

    const: Const

    @property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential const variables contained in this argument."""
        return self.const.unsolved_vars

    def to_hugr(self) -> tys.TypeArg:
        """Computes the Hugr representation of this argument."""
        from guppylang.tys.ty import NumericType

        match self.const:
            case ConstValue(value=v, ty=NumericType(kind=NumericType.Kind.Nat)):
                assert isinstance(v, int)
                return tys.TypeArg(tys.BoundedNatArg(n=v))
            case BoundConstVar(idx=idx):
                hugr_ty = self.const.ty.to_hugr()
                assert isinstance(hugr_ty.root, tys.Opaque)
                param = tys.TypeParam(tys.BoundedNatParam(bound=None))
                return tys.TypeArg(tys.VariableArg(idx=idx, cached_decl=param))
            case ConstValue() | BoundConstVar():
                # TODO: Handle other cases besides nats
                raise NotImplementedError
            case ExistentialConstVar():
                raise InternalGuppyError(
                    "Tried to convert unsolved constant variable to Hugr"
                )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this argument."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> Argument:
        """Accepts a transformer on this argument."""
        return transformer.transform(self) or ConstArg(
            self.const.transform(transformer)
        )



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.prelude import builtins, quantum
from guppylang.prelude.builtins import Bool, Float, Int, List, linst, py
from guppylang.prelude.quantum import qubit



"""Visualise HUGR using graphviz."""

import ast
from collections.abc import Iterable
from typing import TYPE_CHECKING

import graphviz as gv  # type: ignore[import-untyped]

from guppylang.cfg.analysis import (
    DefAssignmentDomain,
    LivenessDomain,
    MaybeAssignmentDomain,
)
from guppylang.cfg.bb import BB, VId
from guppylang.hugr_builder.hugr import DummyOp, Hugr, InPort, Node, OutPort, OutPortV

if TYPE_CHECKING:
    from guppylang.cfg.cfg import CFG

# old palettte: https://colorhunt.co/palette/343a407952b3ffc107e1e8eb
# _COLOURS = {
#     "background": "white",
#     "node": "#7952B3",
#     "edge": "#FFC107",
#     "dark": "#343A40",
#     "const": "#7c55b4",
#     "discard": "#ff8888",
#     "node_border": "#9d80c7",
#     "port_border": "#ffd966",
# }

# ZX colours
# _COLOURS = {
#     "background": "white",
#     "node": "#629DD1",
#     "edge": "#297FD5",
#     "dark": "#112D4E",
#     "const": "#a1eea1",
#     "discard": "#ff8888",
#     "node_border": "#D8F8D8",
#     "port_border": "#E8A5A5",
# }

# Conference talk colours
_COLOURS = {
    "background": "white",
    "node": "#ACCBF9",
    "edge": "#1CADE4",
    "dark": "black",
    "const": "#77CEEF",
    "discard": "#ff8888",
    "node_border": "white",
    "port_border": "#1CADE4",
}


_FONTFACE = "monospace"

_HTML_LABEL_TEMPLATE = """
<TABLE BORDER="{border_width}" CELLBORDER="0" CELLSPACING="1" CELLPADDING="1"
       BGCOLOR="{node_back_color}" COLOR="{border_colour}">
  {inputs_row}
  <TR>
    <TD>
      <TABLE BORDER="0" CELLBORDER="0">
        <TR><TD><FONT POINT-SIZE="{fontsize}" FACE="{fontface}"
            COLOR="{label_color}"><B>{node_label}</B>{node_data}</FONT></TD></TR>
      </TABLE>
    </TD>
  </TR>
  {outputs_row}
</TABLE>
"""


def _format_html_label(**kwargs: str) -> str:
    _HTML_LABEL_DEFAULTS = {
        "label_color": _COLOURS["dark"],
        "node_back_color": _COLOURS["node"],
        "inputs_row": "",
        "outputs_row": "",
        "border_colour": _COLOURS["port_border"],
        "border_width": "1",
        "fontface": _FONTFACE,
        "fontsize": 11.0,
    }
    return _HTML_LABEL_TEMPLATE.format(**{**_HTML_LABEL_DEFAULTS, **kwargs})


_HTML_PORTS_ROW_TEMPLATE = """
    <TR>
        <TD>
            <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="3" CELLPADDING="2">
                <TR>
                    {port_cells}
                </TR>
            </TABLE>
        </TD>
    </TR>
"""

_HTML_PORT_TEMPLATE = (
    '<TD BGCOLOR="{back_colour}" COLOR="{border_colour}"'
    ' PORT="{port_id}" BORDER="{border_width}">'
    '<FONT POINT-SIZE="10.0" FACE="{fontface}" COLOR="{font_colour}">{port}</FONT></TD>'
)

_INPUT_PREFIX = "in."
_OUTPUT_PREFIX = "out."


def _html_ports(ports: Iterable[str], id_prefix: str) -> str:
    return _HTML_PORTS_ROW_TEMPLATE.format(
        port_cells="".join(
            _HTML_PORT_TEMPLATE.format(
                port=port,
                # differentiate input and output node identifiers
                # with a prefix
                port_id=id_prefix + port,
                back_colour=_COLOURS["background"],
                font_colour=_COLOURS["dark"],
                border_width="1",
                border_colour=_COLOURS["port_border"],
                fontface=_FONTFACE,
            )
            for port in ports
        )
    )


def _in_port_name(p: InPort) -> str:
    return (
        f"{p.node.idx}:{_INPUT_PREFIX}{p.offset}"
        if p.offset is not None
        else str(p.node.idx)
    )


def _out_port_name(p: OutPort) -> str:
    return (
        f"{p.node.idx}:{_OUTPUT_PREFIX}{p.offset}"
        if p.offset is not None
        else str(p.node.idx)
    )


def _in_order_name(n: Node) -> str:
    return f"{n.idx}:{_INPUT_PREFIX}None"


def _out_order_name(n: Node) -> str:
    return f"{n.idx}:{_OUTPUT_PREFIX}None"


def viz_node(node: Node, hugr: Hugr, graph: gv.Digraph) -> None:
    in_ports = [str(i) for i in range(node.num_in_ports)]
    out_ports = [str(i) for i in range(node.num_out_ports)]
    if len(node.meta_data) > 0:
        data = "<BR/><BR/>" + "<BR/>".join(
            f"{key}: {value}" for key, value in node.meta_data.items()
        )
    else:
        data = ""
    if len(hugr.children(node)) > 0:
        with graph.subgraph(name=f"cluster{node.idx}") as sub:
            for child in hugr.children(node):
                viz_node(child, hugr, sub)
            html_label = _format_html_label(
                node_back_color=_COLOURS["edge"],
                node_label=node.op.root.display_name(),
                node_data=data,
                border_colour=_COLOURS["port_border"],
                inputs_row=_html_ports(in_ports, _INPUT_PREFIX)
                if len(in_ports) > 0
                else "",
                outputs_row=_html_ports(out_ports, _OUTPUT_PREFIX)
                if len(out_ports) > 0
                else "",
            )
            sub.node(f"{node.idx}", shape="plain", label=f"<{html_label}>")
            sub.attr(label="", margin="10", color=_COLOURS["edge"])
    else:
        html_label = _format_html_label(
            node_back_color=_COLOURS["node"],
            node_label=node.op.name
            if isinstance(node.op, DummyOp)
            else node.op.root.display_name(),
            node_data=data,
            inputs_row=_html_ports(in_ports, _INPUT_PREFIX)
            if len(in_ports) > 0
            else "",
            outputs_row=_html_ports(out_ports, _OUTPUT_PREFIX)
            if len(out_ports) > 0
            else "",
            border_colour=_COLOURS["background"],
        )
        graph.node(f"{node.idx}", label=f"<{html_label}>", shape="plain")


def hugr_to_graphviz(hugr: Hugr) -> gv.Digraph:
    graph_atrr = {
        "rankdir": "",
        "ranksep": "0.1",
        "nodesep": "0.15",
        "margin": "0",
        "bgcolor": _COLOURS["background"],
    }
    graph = gv.Digraph(hugr.name, strict=False)
    graph.attr(**graph_atrr)
    for node in hugr.top_level_nodes():
        viz_node(node, hugr, graph)
    edge_attr = {
        "penwidth": "1.5",
        "arrowhead": "none",
        "arrowsize": "1.0",
        "fontname": _FONTFACE,
        "fontsize": "9",
        "fontcolor": "black",
    }
    for src_port, tgt_port in hugr.edges():
        graph.edge(
            _out_port_name(src_port),
            _in_port_name(tgt_port),
            label=str(src_port.ty) if isinstance(src_port, OutPortV) else "",
            color=_COLOURS["edge"]
            if isinstance(src_port, OutPortV)
            else _COLOURS["dark"],
            **edge_attr,
        )
    for src, tgt in hugr.order_edges():
        graph.edge(
            _out_order_name(src),
            _in_order_name(tgt),
            label="",
            color=_COLOURS["dark"],
            **edge_attr,
        )
    return graph


def render_hugr(hugr: Hugr, filename: str, format_st: str = "svg") -> None:
    gv_graph = hugr_to_graphviz(hugr)
    gv_graph.render(filename, format=format_st)


def commas(*args: str) -> str:
    return ", ".join(args)


def cfg_to_graphviz(
    cfg: "CFG",
    live_before: dict[BB, LivenessDomain[VId]],
    ass_before: dict[BB, DefAssignmentDomain[VId]],
    maybe_ass_before: dict[BB, MaybeAssignmentDomain[VId]],
) -> gv.Digraph:
    graph = gv.Digraph("CFG", strict=False)
    for bb in cfg.bbs:
        label = f"""
assigned: {commas(*(str(x) for x in bb.vars.assigned))}
used: {commas(*(str(x) for x in bb.vars.used))}
maybe_ass_before: {commas(*(str(x) for x in maybe_ass_before[bb]))}
ass_before: {commas(*(str(x) for x in ass_before[bb]))}
live_before: {commas(*(str(x) for x in live_before[bb]))}
--------
""" + "\n".join(ast.unparse(s) for s in bb.statements)
        if bb.branch_pred is not None:
            label += f"\n{ast.unparse(bb.branch_pred)} ?"
        graph.node(str(bb.idx), label, shape="rect")
        for succ in bb.successors:
            graph.edge(str(bb.idx), str(succ.idx))
    return graph


def render_cfg(
    cfg: "CFG",
    live_before: dict[BB, LivenessDomain[VId]],
    ass_before: dict[BB, DefAssignmentDomain[VId]],
    maybe_ass_before: dict[BB, MaybeAssignmentDomain[VId]],
    filename: str,
    format_st: str = "svg",
) -> None:
    gv_graph = cfg_to_graphviz(cfg, live_before, ass_before, maybe_ass_before)
    gv_graph.render(filename, format=format_st)



import ast
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from guppylang.ast_util import AstNode
from guppylang.definition.common import CompiledDef, Definition
from guppylang.error import GuppyError
from guppylang.hugr_builder.hugr import Hugr, OutPortV
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import FunctionType, Type

if TYPE_CHECKING:
    from guppylang.checker.core import Context
    from guppylang.compiler.core import CompiledGlobals, DFContainer


@dataclass(frozen=True)
class ValueDef(Definition):
    """Abstract base class for definitions that represent values."""

    ty: Type

    description: str = field(default="value", init=False)


@dataclass(frozen=True)
class CompiledValueDef(ValueDef, CompiledDef):
    """Abstract base class for compiled definitions that represent values."""

    @abstractmethod
    def load(
        self, dfg: "DFContainer", graph: Hugr, globals: "CompiledGlobals", node: AstNode
    ) -> OutPortV:
        """Loads the defined value into a local Hugr dataflow graph."""


@dataclass(frozen=True)
class CallableDef(ValueDef):
    """Abstract base class for definitions that represent functions."""

    ty: FunctionType

    @abstractmethod
    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""

    @abstractmethod
    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise GuppyError("Guppy functions can only be called in a Guppy context")


class CompiledCallableDef(CallableDef, CompiledValueDef):
    """Abstract base class a global module-level function."""

    ty: FunctionType

    @abstractmethod
    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: "DFContainer",
        graph: Hugr,
        globals: "CompiledGlobals",
        node: AstNode,
    ) -> list[OutPortV]:
        """Compiles a call to the function."""

    @abstractmethod
    def load_with_args(
        self,
        type_args: Inst,
        dfg: "DFContainer",
        graph: Hugr,
        globals: "CompiledGlobals",
        node: AstNode,
    ) -> OutPortV:
        """Loads the function into a local Hugr dataflow graph.

        Requires an instantiation for all function parameters.
        """

    def load(
        self, dfg: "DFContainer", graph: Hugr, globals: "CompiledGlobals", node: AstNode
    ) -> OutPortV:
        """Loads the defined value into a local Hugr dataflow graph."""
        return self.load_with_args([], dfg, graph, globals, node)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias

from hugr.serialization import tys
from hugr.serialization.tys import TypeBound
from typing_extensions import Self

from guppylang.ast_util import AstNode
from guppylang.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.common import ToHugr
from guppylang.tys.const import BoundConstVar, ExistentialConstVar
from guppylang.tys.var import ExistentialVar

if TYPE_CHECKING:
    from guppylang.tys.ty import Type


# We define the `Parameter` type as a union of all `ParameterBase` subclasses defined
# below. This models an algebraic data type and enables exhaustiveness checking in
# pattern matches etc.
# Note that this might become obsolete in case the `@sealed` decorator is added:
#  * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#  * https://github.com/johnthagen/sealed-typing-pep
Parameter: TypeAlias = "TypeParam | ConstParam"


@dataclass(frozen=True)
class ParameterBase(ToHugr[tys.TypeParam], ABC):
    """Abstract base class for parameters used in function and type definitions.

    For example, when defining a struct type

    ```
    @guppy.struct
    class Foo[T, n: int]:
        ...
    ```

    we generate an `StructDef` that depends on the parameters `T` and `n`. From
    this, we obtain a proper `StructType` by providing arguments that are substituted
    for the parameters (for example `Foo[int, 42]`).
    """

    idx: int
    name: str

    @abstractmethod
    def with_idx(self, idx: int) -> Self:
        """Returns a copy of the parameter with a new index."""

    @abstractmethod
    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> Argument:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """

    @abstractmethod
    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """

    @abstractmethod
    def to_bound(self, idx: int | None = None) -> Argument:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """


@dataclass(frozen=True)
class TypeParam(ParameterBase):
    """A parameter of kind type. Used to define generic functions and types."""

    can_be_linear: bool

    def with_idx(self, idx: int) -> "TypeParam":
        """Returns a copy of the parameter with a new index."""
        return TypeParam(idx, self.name, self.can_be_linear)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> TypeArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                raise GuppyTypeError(
                    f"Expected a type, got value of type {const.ty}", loc
                )
            case TypeArg(ty):
                if not self.can_be_linear and ty.linear:
                    raise GuppyTypeError(
                        f"Expected a non-linear type, got value of type {ty}", loc
                    )
                return arg

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        from guppylang.tys.ty import ExistentialTypeVar

        var = ExistentialTypeVar.fresh(self.name, self.can_be_linear)
        return TypeArg(var), var

    def to_bound(self, idx: int | None = None) -> Argument:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        from guppylang.tys.ty import BoundTypeVar

        if idx is None:
            idx = self.idx
        return TypeArg(BoundTypeVar(self.name, idx, self.can_be_linear))

    def to_hugr(self) -> tys.TypeParam:
        """Computes the Hugr representation of the parameter."""
        return tys.TypeParam(
            tys.TypeTypeParam(
                b=tys.TypeBound.Any if self.can_be_linear else TypeBound.Copyable
            )
        )


@dataclass(frozen=True)
class ConstParam(ParameterBase):
    """A parameter of kind constant. Used to define fixed-size arrays etc."""

    ty: "Type"

    def __post_init__(self) -> None:
        if self.ty.unsolved_vars:
            raise InternalGuppyError(
                "Attempted to create constant param with unsolved type"
            )

    def with_idx(self, idx: int) -> "ConstParam":
        """Returns a copy of the parameter with a new index."""
        return ConstParam(idx, self.name, self.ty)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> ConstArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                if const.ty != self.ty:
                    raise GuppyTypeError(
                        f"Expected argument of type `{self.ty}`, got {const.ty}", loc
                    )
                return arg
            case TypeArg():
                raise GuppyTypeError(
                    f"Expected argument of type `{self.ty}`, got type", loc
                )

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        var = ExistentialConstVar.fresh(self.name, self.ty)
        return ConstArg(var), var

    def to_bound(self, idx: int | None = None) -> Argument:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        if idx is None:
            idx = self.idx
        return ConstArg(BoundConstVar(self.ty, self.name, idx))

    def to_hugr(self) -> tys.TypeParam:
        """Computes the Hugr representation of the parameter."""
        from guppylang.tys.ty import NumericType

        match self.ty:
            case NumericType(kind=NumericType.Kind.Nat):
                return tys.TypeParam(tys.BoundedNatParam(bound=None))
            case _:
                hugr_ty = self.ty.to_hugr()
                assert isinstance(hugr_ty.root, tys.Opaque)
                return tys.TypeParam(tys.StringParam())


def check_all_args(
    params: Sequence[Parameter],
    args: Sequence[Argument],
    type_name: str,
    loc: AstNode | None = None,
) -> None:
    """Checks a list of arguments against the given parameters.

    Raises a user error if number of arguments doesn't match or one of the argument is
    invalid.
    """
    exp, act = len(params), len(args)
    if exp > act:
        raise GuppyError(f"Missing parameter for type `{type_name}`", loc)
    elif 0 == exp < act:
        raise GuppyError(f"Type `{type_name}` is not parameterized", loc)
    elif 0 < exp < act:
        raise GuppyError(f"Too many parameters for type `{type_name}`", loc)

    # Now check that the kinds match up
    for param, arg in zip(params, args, strict=True):
        # TODO: The error location is bad. We want the location of `arg`, not of the
        #  whole thing.
        param.check_arg(arg, loc)



from abc import abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from hugr.serialization import tys

from guppylang.ast_util import AstNode
from guppylang.definition.common import CompiledDef, Definition
from guppylang.tys.arg import Argument
from guppylang.tys.param import Parameter, check_all_args
from guppylang.tys.ty import OpaqueType, Type

if TYPE_CHECKING:
    from guppylang.checker.core import Globals


@dataclass(frozen=True)
class TypeDef(Definition):
    """Abstract base class for type definitions."""

    description: str = field(default="type", init=False)

    @abstractmethod
    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> Type:
        """Checks if the type definition can be instantiated with the given arguments.

        Returns the resulting concrete type or raises a user error if the arguments are
        invalid.
        """


@dataclass(frozen=True)
class OpaqueTypeDef(TypeDef, CompiledDef):
    """An opaque type definition that is backed by some Hugr type."""

    params: Sequence[Parameter]
    always_linear: bool
    to_hugr: Callable[[Sequence[Argument]], tys.Type]
    bound: tys.TypeBound | None = None

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> OpaqueType:
        """Checks if the type definition can be instantiated with the given arguments.

        Returns the resulting concrete type or raises a user error if the arguments are
        invalid.
        """
        check_all_args(self.params, args, self.name, loc)
        return OpaqueType(args, self)



"""Linearity checking

Linearity checking across control-flow is done by the `CFGChecker`.
"""

import ast
from collections.abc import Generator, Iterator
from contextlib import contextmanager

from guppylang.ast_util import AstNode, find_nodes, get_type
from guppylang.cfg.analysis import LivenessAnalysis
from guppylang.cfg.bb import BB, VariableStats
from guppylang.checker.cfg_checker import CheckedBB, CheckedCFG, Signature
from guppylang.checker.core import (
    FieldAccess,
    Locals,
    Place,
    PlaceId,
    Variable,
)
from guppylang.error import GuppyError, GuppyTypeError
from guppylang.nodes import (
    CheckedNestedFunctionDef,
    DesugaredGenerator,
    DesugaredListComp,
    FieldAccessAndDrop,
    PlaceNode,
)
from guppylang.tys.ty import StructType


class Scope(Locals[PlaceId, Place]):
    """Scoped collection of assigned places indexed by their id.

    Keeps track of which places have already been used.
    """

    parent_scope: "Scope | None"
    used_local: dict[PlaceId, AstNode]
    used_parent: dict[PlaceId, AstNode]

    def __init__(self, parent: "Scope | None" = None):
        self.used_local = {}
        self.used_parent = {}
        super().__init__({}, parent)

    def used(self, x: PlaceId) -> AstNode | None:
        """Checks whether a place has already been used."""
        if x in self.vars:
            return self.used_local.get(x, None)
        assert self.parent_scope is not None
        return self.parent_scope.used(x)

    def use(self, x: PlaceId, node: AstNode) -> None:
        """Records a use of a place.

        Works for places in the current scope as well as places in any parent scope.
        """
        if x in self.vars:
            self.used_local[x] = node
        else:
            assert self.parent_scope is not None
            assert x in self.parent_scope
            self.used_parent[x] = node
            self.parent_scope.use(x, node)

    def assign(self, place: Place) -> None:
        """Records an assignment of a place."""
        assert place.defined_at is not None
        x = place.id
        self.vars[x] = place
        if x in self.used_local:
            self.used_local.pop(x)

    def stats(self) -> VariableStats[PlaceId]:
        assigned = {}
        for x, place in self.vars.items():
            assert place.defined_at is not None
            assigned[x] = place.defined_at
        return VariableStats(assigned, self.used_parent)


class BBLinearityChecker(ast.NodeVisitor):
    """AST visitor that checks linearity for a single basic block."""

    scope: Scope
    stats: VariableStats[PlaceId]

    def check(self, bb: "CheckedBB[Variable]", is_entry: bool) -> Scope:
        # Manufacture a scope that holds all places that are live at the start
        # of this BB
        input_scope = Scope()
        for var in bb.sig.input_row:
            for place in leaf_places(var):
                input_scope.assign(place)

        # Open up a new nested scope to check the BB contents. This way we can track
        # when we use variables from the outside vs ones assigned in this BB. The only
        # exception is the entry BB since function arguments should be treated as part
        # of the entry BB
        self.scope = input_scope if is_entry else Scope(input_scope)

        for stmt in bb.statements:
            self.visit(stmt)
        if bb.branch_pred:
            self.visit(bb.branch_pred)
        return self.scope

    @contextmanager
    def new_scope(self) -> Generator[Scope, None, None]:
        scope, new_scope = self.scope, Scope(self.scope)
        self.scope = new_scope
        yield new_scope
        self.scope = scope

    def visit_PlaceNode(self, node: PlaceNode) -> None:
        for place in leaf_places(node.place):
            x = place.id
            if (use := self.scope.used(x)) and place.ty.linear:
                raise GuppyError(
                    f"{place.describe} with linear type `{place.ty}` was already used "
                    "(at {0})",
                    node,
                    [use],
                )
            self.scope.use(x, node)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit(node.value)
        self._check_assign_targets(node.targets)

    def visit_FieldAccessAndDrop(self, node: FieldAccessAndDrop) -> None:
        # A field access on a value that is not a place. This means the value can no
        # longer be accessed after the field has been projected out. Thus, this is only
        # legal if there are no remaining linear fields on the value
        self.visit(node.value)
        for field in node.struct_ty.fields:
            if field.name != node.field.name and field.ty.linear:
                raise GuppyTypeError(
                    f"Linear field `{field.name}` of expression with type "
                    f"`{node.struct_ty}` is not used",
                    node.value,
                )

    def visit_Expr(self, node: ast.Expr) -> None:
        # An expression statement where the return value is discarded
        self.visit(node.value)
        ty = get_type(node.value)
        if ty.linear:
            raise GuppyTypeError(f"Value with linear type `{ty}` is not used", node)

    def visit_DesugaredListComp(self, node: DesugaredListComp) -> None:
        self._check_comprehension(node, node.generators)

    def visit_CheckedNestedFunctionDef(self, node: CheckedNestedFunctionDef) -> None:
        # Linearity of the nested function has already been checked. We just need to
        # verify that no linear variables are captured
        # TODO: In the future, we could support capturing of non-linear subplaces
        for var, use in node.captured.values():
            if var.ty.linear:
                raise GuppyError(
                    f"{var.describe} with linear type `{var.ty}` may not be used here "
                    f"because it was defined in an outer scope (at {{0}})",
                    use,
                    [var.defined_at],
                )
            for place in leaf_places(var):
                self.scope.use(place.id, use)
        self.scope.assign(Variable(node.name, node.ty, node))

    def _check_assign_targets(self, targets: list[ast.expr]) -> None:
        """Helper function to check assignments."""
        # We're not allowed to override an unused linear place
        [target] = targets
        for tgt in find_nodes(lambda n: isinstance(n, PlaceNode), target):
            assert isinstance(tgt, PlaceNode)
            for tgt_place in leaf_places(tgt.place):
                x = tgt_place.id
                # Only check for overrides of places locally defined in this BB. Global
                # checks are handled by dataflow analysis.
                if x in self.scope.vars and x not in self.scope.used_local:
                    place = self.scope[x]
                    if place.ty.linear:
                        raise GuppyError(
                            f"{place.describe} with linear type `{place.ty}` is not "
                            "used",
                            place.defined_at,
                        )
                self.scope.assign(tgt_place)

    def _check_comprehension(
        self, node: DesugaredListComp, gens: list[DesugaredGenerator]
    ) -> None:
        """Helper function to recursively check list comprehensions."""
        if not gens:
            self.visit(node.elt)
            return

        # Check the iterator expression in the current scope
        gen, *gens = gens
        self.visit(gen.iter_assign.value)
        assert isinstance(gen.iter, PlaceNode)

        # The rest is checked in a new nested scope so we can track which variables
        # are introduced and used inside the loop
        with self.new_scope() as inner_scope:
            # In particular, assign the iterator variable in the new scope
            self._check_assign_targets(gen.iter_assign.targets)
            self.visit(gen.hasnext_assign)
            self.visit(gen.next_assign)

            # `if` guards are generally not allowed when we're iterating over linear
            # variables. The only exception is if all linear variables are already
            # consumed by the first guard
            if gen.ifs:
                first_if, *other_ifs = gen.ifs
                # Check if there are linear iteration variables that have not been used
                # by the first guard
                self.visit(first_if)
                for place in self.scope.vars.values():
                    # The only exception is the iterator variable since we make sure
                    # that it is carried through each iteration during Hugr generation
                    if place == gen.iter.place:
                        continue
                    for leaf in leaf_places(place):
                        x = leaf.id
                        if not self.scope.used(x) and place.ty.linear:
                            raise GuppyTypeError(
                                f"{place.describe} with linear type `{place.ty}` is "
                                "not used on all control-flow paths of the list "
                                "comprehension",
                                place.defined_at,
                            )
                for expr in other_ifs:
                    self.visit(expr)

            # Recursively check the remaining generators
            self._check_comprehension(node, gens)

            # Check the iter finalizer so we record a final use of the iterator
            self.visit(gen.iterend)

            # We have to make sure that all linear variables that were introduced in the
            # inner scope have been used
            for place in inner_scope.vars.values():
                for leaf in leaf_places(place):
                    x = leaf.id
                    if leaf.ty.linear and not inner_scope.used(x):
                        raise GuppyTypeError(
                            f"{leaf.describe} with linear type `{leaf.ty}` is not used",
                            leaf.defined_at,
                        )

            # On the other hand, we have to ensure that no linear places from the
            # outer scope have been used inside the comprehension (they would be used
            # multiple times since the comprehension body is executed repeatedly)
            for x, use in inner_scope.used_parent.items():
                place = inner_scope[x]
                if place.ty.linear:
                    raise GuppyTypeError(
                        f"{place.describe} with linear type `{place.ty}` would be used "
                        "multiple times when evaluating this comprehension",
                        use,
                    )


def leaf_places(place: Place) -> Iterator[Place]:
    """Returns all leaf descendant projections of a place."""
    stack = [place]
    while stack:
        place = stack.pop()
        if isinstance(place.ty, StructType):
            stack += [
                FieldAccess(place, field, place.defined_at) for field in place.ty.fields
            ]
        else:
            yield place


def check_cfg_linearity(cfg: "CheckedCFG[Variable]") -> "CheckedCFG[Place]":
    """Checks whether a CFG satisfies the linearity requirements.

    Raises a user-error if linearity violations are found.

    Returns a new CFG with refined basic block signatures in terms of *places* rather
    than just variables.
    """
    bb_checker = BBLinearityChecker()
    scopes: dict[BB, Scope] = {
        bb: bb_checker.check(bb, is_entry=bb == cfg.entry_bb) for bb in cfg.bbs
    }

    # Run liveness analysis
    stats = {bb: scope.stats() for bb, scope in scopes.items()}
    live_before = LivenessAnalysis(stats).run(cfg.bbs)

    # Construct a CFG that tracks places instead of just variables
    result_cfg: CheckedCFG[Place] = CheckedCFG(cfg.input_tys, cfg.output_ty)
    checked: dict[BB, CheckedBB[Place]] = {}

    for bb, scope in scopes.items():
        # We have to check that used linear variables are not being outputted
        for succ in bb.successors:
            live = live_before[succ]
            for x, use_bb in live.items():
                use_scope = scopes[use_bb]
                place = use_scope[x]
                if place.ty.linear and (use := scope.used(x)):
                    raise GuppyError(
                        f"{place.describe} with linear type `{place.ty}` was "
                        "already used (at {0})",
                        use_scope.used_parent[x],
                        [use],
                    )

            # On the other hand, unused linear variables *must* be outputted
            for place in scope.values():
                for leaf in leaf_places(place):
                    x = leaf.id
                    used_later = x in live
                    if leaf.ty.linear and not scope.used(x) and not used_later:
                        # TODO: This should be "Variable x with linear type ty is not
                        #  used in {bb}". But for this we need a way to associate BBs
                        #  with source locations.
                        raise GuppyError(
                            f"{leaf.describe} with linear type `{leaf.ty}` is "
                            "not used on all control-flow paths",
                            # Re-lookup defined_at in scope because we might have a
                            # more precise location
                            scope[x].defined_at,
                        )

        assert isinstance(bb, CheckedBB)
        sig = Signature(
            input_row=[scope[x] for x in live_before[bb]]
            if bb not in (cfg.entry_bb, cfg.exit_bb)
            else bb.sig.input_row,
            output_rows=[
                [scope[x] for x in live_before[succ]] for succ in bb.successors
            ],
        )
        checked[bb] = CheckedBB(
            bb.idx, result_cfg, bb.statements, branch_pred=bb.branch_pred, sig=sig
        )

    # Fill in missing fields of the result CFG
    result_cfg.bbs = list(checked.values())
    result_cfg.entry_bb = checked[cfg.entry_bb]
    result_cfg.exit_bb = checked[cfg.exit_bb]
    result_cfg.live_before = {checked[bb]: cfg.live_before[bb] for bb in cfg.bbs}
    result_cfg.ass_before = {checked[bb]: cfg.ass_before[bb] for bb in cfg.bbs}
    result_cfg.maybe_ass_before = {
        checked[bb]: cfg.maybe_ass_before[bb] for bb in cfg.bbs
    }
    for bb in cfg.bbs:
        checked[bb].predecessors = [checked[pred] for pred in bb.predecessors]
        checked[bb].successors = [checked[succ] for succ in bb.successors]
    return result_cfg



import itertools
from abc import ABC
from collections.abc import Iterator
from dataclasses import dataclass
from typing import ClassVar

# Type of de Bruijn indices
DeBruijn = int

# Type of unique variable identifiers
UniqueId = int


@dataclass(frozen=True)
class Var(ABC):
    """Abstract base class for variables that occur in types.

    A variable can either occur as a type itself (see subclasses `BoundTypeVar` and
    `ExistentialTypeVar`) or as an argument to a parametrised type.
    """

    # Name that is used when showing the variable to the user
    display_name: str


@dataclass(frozen=True)
class BoundVar(Var, ABC):
    """Variable that is bound to a parameter of kind `Const`.

    Identified by a de Bruijn index.
    """

    idx: DeBruijn


@dataclass(frozen=True)
class ExistentialVar(Var, ABC):
    """Existential variable, referencing a parameter of kind `Const`.

    Identified by a globally unique id.

    During type checking we try to solve all existential variables and substitute
    them with concrete consts.
    """

    id: UniqueId

    # Generator of fresh unique ids
    _fresh_id: ClassVar[Iterator[UniqueId]] = itertools.count()



import itertools
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, Optional

import networkx as nx  # type: ignore[import-untyped]
from hugr import node_port
from hugr.serialization import ops, tys
from hugr.serialization import serial_hugr as raw
from hugr.serialization.ops import OpType

from guppylang.tys.subst import Inst
from guppylang.tys.ty import (
    FunctionType,
    StructType,
    SumType,
    TupleType,
    Type,
    TypeRow,
    row_to_type,
    rows_to_hugr,
    type_to_row,
)

NodeIdx = int
PortOffset = int


@dataclass(frozen=True)
class Port(ABC):
    """Base class for ports on nodes."""

    node: "Node"
    offset: PortOffset | None


class InPort(Port, ABC):
    """Base class for a port that incoming wires connect to."""


class OutPort(Port, ABC):
    """Base class for a port that outgoing wires come from."""


@dataclass(frozen=True)
class InPortV(InPort):
    """A typed value input port."""

    ty: Type
    offset: PortOffset


@dataclass(frozen=True)
class OutPortV(OutPort):
    """A typed value output port."""

    ty: Type
    offset: PortOffset


@dataclass(frozen=True)
class InPortCF(InPort):
    """A control-flow input port."""

    # Control flow inputs are unordered so no port offset is needed
    offset: None = field(default=None, init=False)


class OutPortCF(OutPort):
    """A control-flow output port."""


class DummyOp(OpType):
    """A dummy Hugr op that is replaced with a call to a dummy function declaration
    during serialisation.

    This is a placeholder for ops that aren't yet available in Hugr.
    """

    root: str  # type: ignore[assignment]

    @property
    def name(self) -> str:
        """The name of the dummy op."""
        return self.root


Edge = tuple[OutPort, InPort]

TypeList = list[Type]


@dataclass
class Node(ABC):
    """Base class for a node in the graph.

    Has a number of input and output ports and an associated op type.
    """

    idx: NodeIdx
    op: ops.OpType
    parent: Optional["Node"]
    meta_data: dict[str, Any]

    @property
    @abstractmethod
    def num_in_ports(self) -> int:
        """The number of input ports on this node."""

    @property
    @abstractmethod
    def num_out_ports(self) -> int:
        """The number of output ports on this node."""

    @abstractmethod
    def in_port(self, offset: PortOffset | None) -> InPort:
        """Returns the input port at the given offset."""

    @abstractmethod
    def out_port(self, offset: PortOffset | None) -> OutPort:
        """Returns the output port at the given offset."""

    def update_op(self) -> None:  # noqa: B027
        """Updates the op type associated with this node with additional information.

        This should be called before serialisation.
        """

    @property
    def in_ports(self) -> Iterator[InPort]:
        """Returns an iterator over all input ports from left to right."""
        return (self.in_port(i) for i in range(self.num_in_ports))

    @property
    def out_ports(self) -> Iterator[OutPort]:
        """Returns an iterator over all output ports from left to right."""
        return (self.out_port(i) for i in range(self.num_out_ports))


@dataclass
class VNode(Node):
    """A node with typed value ports."""

    in_port_types: TypeList
    out_port_types: TypeList

    @property
    def num_in_ports(self) -> int:
        """The number of input ports on this node."""
        return len(self.in_port_types)

    @property
    def num_out_ports(self) -> int:
        """The number of output ports on this node."""
        return len(self.out_port_types)

    def add_in_port(self, ty: Type) -> InPortV:
        """Adds an input port at the end of the node and returns the port."""
        p = InPortV(self, self.num_in_ports, ty)
        self.in_port_types.append(ty)
        return p

    def add_out_port(self, ty: Type) -> OutPortV:
        """Adds an output port at the end of the node and returns the port."""
        p = OutPortV(self, self.num_out_ports, ty)
        self.out_port_types.append(ty)
        return p

    def in_port(self, offset: PortOffset | None) -> InPortV:
        """Returns the input port at the given offset."""
        assert offset is not None
        assert offset < self.num_in_ports
        assert offset != -1, "Cannot get the port of an order edge"
        return InPortV(self, offset, self.in_port_types[offset])

    def out_port(self, offset: PortOffset | None) -> OutPortV:
        """Returns the output port at the given offset."""
        assert offset is not None
        assert offset < self.num_out_ports
        assert offset != -1, "Cannot get the port of an order edge"
        return OutPortV(self, offset, self.out_port_types[offset])

    @property
    def in_ports(self) -> Iterator[InPortV]:
        """Returns an iterator over all input ports from left to right."""
        return (self.in_port(i) for i in range(self.num_in_ports))

    @property
    def out_ports(self) -> Iterator[OutPortV]:
        """Returns an iterator over all output ports from left to right."""
        return (self.out_port(i) for i in range(self.num_out_ports))

    def update_op(self) -> None:
        """Updates the operation associated with this node with type information.

        Feeds type information from the in- and out-ports to the operation class to
        update signature information. This function must be called before serialisation.
        """
        # We can't call `to_hugr()` on polymorphic function types, so we have to skip
        # ops that have connected `Function` edges.
        has_poly_func_edge = isinstance(
            self.op.root, ops.FuncDecl | ops.FuncDefn | ops.Call | ops.LoadFunction
        )
        if isinstance(self.op.root, ops.BaseOp) and not has_poly_func_edge:
            in_types = [t.to_hugr() for t in self.in_port_types]
            out_types = [t.to_hugr() for t in self.out_port_types]
            self.op.root.insert_port_types(in_types, out_types)
        super().update_op()


class CFNode(Node):
    """A node in a control-flow graph.

    Compared to value nodes, the ports on this node are not typed since they correspond
    to control-flow instead of data-flow.
    """

    _num_out_ports: int = 0

    @property
    def num_in_ports(self) -> int:
        """The number of input ports on this node."""
        return 0

    @property
    def num_out_ports(self) -> int:
        """The number of output ports on this node."""
        return self._num_out_ports

    def add_out_port(self) -> OutPortCF:
        """Adds an output port at the end of the node and returns the port."""
        p = OutPortCF(self, self.num_out_ports)
        self._num_out_ports += 1
        return p

    def in_port(self, offset: PortOffset | None) -> InPortCF:
        assert offset is None
        return InPortCF(self)

    def out_port(self, offset: PortOffset | None) -> OutPortCF:
        """Returns the output port at the given offset."""
        assert offset is not None
        assert offset < self.num_out_ports
        return OutPortCF(self, offset)

    def update_op(self) -> None:
        super().update_op()


class DFContainingNode(Node, ABC):
    """Base class for a node whose children form a dataflow graph.

    Compared to a normal node, this node tracks the `Input` and `Output` nodes of its
    child DFG which is required to compute the operation signature.
    """

    input_child: Optional["VNode"] = None  # Input Node for the child dataflow graph
    output_child: Optional["VNode"] = None  # Output Node for the child dataflow graph

    def update_op(self) -> None:
        """Updates the operation associated with this node with type information.

        Feeds type information from the signature of the contained dataflow graph to
        the operation class to. This function must be called before serialisation.
        """
        assert self.input_child is not None
        assert self.output_child is not None
        assert isinstance(self.op.root, ops.BaseOp)
        # Input and output node may have extra order edges connected, so we filter
        # `None`s here
        ins = [ty.to_hugr() for ty in self.input_child.out_port_types]
        outs = [ty.to_hugr() for ty in self.output_child.in_port_types]
        self.op.root.insert_child_dfg_signature(inputs=ins, outputs=outs)
        super().update_op()


class DFContainingVNode(VNode, DFContainingNode):
    """A value node whose children form a dataflow graph"""


class BlockNode(DFContainingNode, CFNode):
    """A `Block` node representing a basic block."""


OrderEdge = tuple["Node", "Node"]
ORDER_EDGE_KEY = (-1, -1)

UNDEFINED: node_port.NodeIdx = -1


class Hugr:
    """Hierarchical unified graph representation."""

    name: str
    root: VNode
    _graph: nx.MultiDiGraph  # TODO: We probably don't need networkx.
    _children: dict[NodeIdx, list[Node]]
    _default_parent: Node | None

    def __init__(self, name: str | None = None) -> None:
        """Creates a new Hugr."""
        self.name = name or "Unnamed"
        self._default_parent = None
        self._graph = nx.MultiDiGraph()
        self._children = {-1: []}
        self.root = self.add_node(
            op=ops.OpType(ops.Module(parent=UNDEFINED)),
            meta_data={"name": name},
            parent=None,
        )

    @contextmanager
    def parent(self, parent: Node) -> Iterator[None]:
        """Context manager to set a default parent for adding new nodes."""
        old_default = self._default_parent
        self._default_parent = parent
        yield
        self._default_parent = old_default

    def _insert_node(self, node: Node, inputs: list[OutPortV] | None = None) -> None:
        """Helper method to insert a node into the graph datastructure."""
        self._graph.add_node(node.idx, data=node)
        self._children[node.idx] = []
        self._children[node.parent.idx if node.parent else -1].append(node)
        if inputs is not None:
            for i, port in enumerate(inputs):
                self.add_edge(port, node.in_port(i))

    def add_node(
        self,
        op: ops.OpType,
        input_types: TypeList | None = None,
        output_types: TypeList | None = None,
        parent: Node | None = None,
        inputs: list[OutPortV] | None = None,
        meta_data: dict[str, Any] | None = None,
    ) -> VNode:
        """Helper method to add a generic value node to the graph."""
        input_types = input_types or []
        output_types = output_types or []
        parent = parent or self._default_parent
        node = VNode(
            idx=self._graph.number_of_nodes(),
            op=op,
            parent=parent,
            in_port_types=[p.ty for p in inputs] if inputs is not None else input_types,
            out_port_types=output_types,
            meta_data=meta_data or {},
        )
        self._insert_node(node, inputs)
        return node

    def _add_dfg_node(
        self,
        op: ops.OpType,
        input_types: TypeList | None = None,
        output_types: TypeList | None = None,
        parent: Node | None = None,
        inputs: list[OutPortV] | None = None,
        meta_data: dict[str, Any] | None = None,
    ) -> DFContainingVNode:
        """Helper method to add a generic dataflow containing value node to the
        graph."""
        input_types = input_types or []
        output_types = output_types or []
        parent = parent or self._default_parent
        node = DFContainingVNode(
            idx=self._graph.number_of_nodes(),
            op=op,
            parent=parent,
            in_port_types=[p.ty for p in inputs] if inputs is not None else input_types,
            out_port_types=output_types,
            meta_data=meta_data or {},
        )
        self._insert_node(node, inputs)
        return node

    def set_root_name(self, name: str) -> VNode:
        """Sets the name of the root node."""
        self.root.meta_data["name"] = name
        return self.root

    def add_constant(
        self, value: ops.Value, ty: Type, parent: Node | None = None
    ) -> VNode:
        """Adds a constant node holding a given value to the graph."""
        return self.add_node(
            ops.OpType(ops.Const(v=value, parent=UNDEFINED)), [], [ty], parent, None
        )

    def add_input(
        self, output_tys: TypeList | None = None, parent: Node | None = None
    ) -> VNode:
        """Adds an `Input` node to the graph."""
        parent = parent or self._default_parent
        node = self.add_node(
            ops.OpType(ops.Input(parent=UNDEFINED)), [], output_tys, parent
        )
        if isinstance(parent, DFContainingNode):
            parent.input_child = node
        return node

    def add_input_with_ports(
        self, output_tys: Sequence[Type], parent: Node | None = None
    ) -> tuple[VNode, list[OutPortV]]:
        """Adds an `Input` node to the graph."""
        node = self.add_input(None, parent)
        ports = [node.add_out_port(ty) for ty in output_tys]
        return node, ports

    def add_output(
        self,
        inputs: list[OutPortV] | None = None,
        input_tys: TypeList | None = None,
        parent: Node | None = None,
    ) -> VNode:
        """Adds an `Output` node to the graph."""
        parent = parent or self._default_parent
        node = self.add_node(
            ops.OpType(ops.Output(parent=UNDEFINED)), input_tys, [], parent, inputs
        )
        if isinstance(parent, DFContainingNode):
            parent.output_child = node
        return node

    def add_block(self, parent: Node | None, num_successors: int = 0) -> BlockNode:
        """Adds a `Block` node to the graph."""
        node = BlockNode(
            idx=self._graph.number_of_nodes(),
            op=ops.OpType(ops.DataflowBlock(parent=UNDEFINED, sum_rows=[])),
            parent=parent,
            meta_data={},
        )
        self._insert_node(node)
        for _ in range(num_successors):
            node.add_out_port()
        return node

    def add_exit(self, output_tys: TypeList, parent: Node) -> CFNode:
        """Adds an `Exit` node to the graph."""
        outputs = [ty.to_hugr() for ty in output_tys]
        node = CFNode(
            idx=self._graph.number_of_nodes(),
            op=ops.OpType(ops.ExitBlock(cfg_outputs=outputs, parent=UNDEFINED)),
            parent=parent,
            meta_data={},
        )
        self._insert_node(node)
        return node

    def add_dfg(self, parent: Node) -> DFContainingVNode:
        """Adds a nested dataflow `DFG` node to the graph."""
        return self._add_dfg_node(ops.OpType(ops.DFG(parent=UNDEFINED)), [], [], parent)

    def add_case(self, parent: Node) -> DFContainingVNode:
        """Adds a `Case` node to the graph."""
        return self._add_dfg_node(
            ops.OpType(ops.Case(parent=UNDEFINED)), [], [], parent
        )

    def add_cfg(self, parent: Node, inputs: list[OutPortV]) -> VNode:
        """Adds a nested control-flow `CFG` node to the graph."""
        return self.add_node(
            ops.OpType(ops.CFG(parent=UNDEFINED)), [], [], parent, inputs
        )

    def add_conditional(
        self,
        cond_input: OutPortV,
        inputs: list[OutPortV],
        parent: Node | None = None,
    ) -> VNode:
        """Adds a `Conditional` node to the graph."""
        inputs = [cond_input, *inputs]
        return self.add_node(
            ops.OpType(ops.Conditional(sum_rows=[], parent=UNDEFINED)),
            None,
            None,
            parent,
            inputs,
        )

    def add_tail_loop(
        self, inputs: list[OutPortV], parent: Node | None = None
    ) -> DFContainingVNode:
        """Adds a `TailLoop` node to the graph."""
        return self._add_dfg_node(
            ops.OpType(ops.TailLoop(parent=UNDEFINED)), None, None, parent, inputs
        )

    def add_make_tuple(
        self, inputs: list[OutPortV], parent: Node | None = None
    ) -> VNode:
        """Adds a `MakeTuple` node to the graph."""
        ty = TupleType([port.ty for port in inputs])
        return self.add_node(
            ops.OpType(ops.MakeTuple(parent=UNDEFINED)), None, [ty], parent, inputs
        )

    def add_unpack_tuple(
        self, input_tuple: OutPortV, parent: Node | None = None
    ) -> VNode:
        """Adds an `UnpackTuple` node to the graph."""
        match input_tuple.ty:
            case TupleType(element_types=elems):
                tys = list(elems)
            case StructType(fields=fields):
                tys = [field.ty for field in fields]
            case ty:
                raise AssertionError(f"Cannot unpack `{ty}`")
        return self.add_node(
            ops.OpType(ops.UnpackTuple(parent=UNDEFINED)),
            None,
            tys,
            parent,
            [input_tuple],
        )

    def add_tag(
        self,
        variants: Sequence[TypeRow],
        tag: int,
        inputs: list[OutPortV],
        parent: Node | None = None,
    ) -> VNode:
        """Adds a `Tag` node to the graph."""
        assert all(inp.ty == ty for inp, ty in zip(inputs, variants[tag], strict=True))
        hugr_variants = rows_to_hugr(variants)
        out_ty = SumType([row_to_type(row) for row in variants])
        return self.add_node(
            ops.OpType(ops.Tag(tag=tag, variants=hugr_variants, parent=UNDEFINED)),
            None,
            [out_ty],
            parent,
            inputs,
        )

    def add_load_constant(
        self, const_port: OutPortV, parent: Node | None = None
    ) -> VNode:
        """Adds a `LoadConstant` node to the graph."""
        return self.add_node(
            ops.OpType(
                ops.LoadConstant(datatype=const_port.ty.to_hugr(), parent=UNDEFINED)
            ),
            None,
            [const_port.ty],
            parent,
            [const_port],
        )

    def add_load_function(
        self, def_port: OutPortV, inst: Inst, parent: Node | None = None
    ) -> VNode:
        """Adds a `LoadFunction` node to the graph."""
        assert isinstance(def_port.ty, FunctionType)
        assert len(def_port.ty.params) == len(inst)
        instantiation = def_port.ty.instantiate(inst)
        op = ops.LoadFunction(
            func_sig=def_port.ty.to_hugr_poly(),
            type_args=[arg.to_hugr() for arg in inst],
            signature=tys.FunctionType(input=[], output=[instantiation.to_hugr()]),
            parent=UNDEFINED,
        )
        return self.add_node(ops.OpType(op), None, [instantiation], parent, [def_port])

    def add_call(
        self,
        def_port: OutPortV,
        args: list[OutPortV],
        inst: Inst,
        parent: Node | None = None,
    ) -> VNode:
        """Adds a `Call` node to the graph."""
        assert isinstance(def_port.ty, FunctionType)
        instantiation = def_port.ty.instantiate(inst)
        op = ops.Call(
            func_sig=def_port.ty.to_hugr_poly(),
            type_args=[arg.to_hugr() for arg in inst],
            instantiation=instantiation.to_hugr().root,
            parent=UNDEFINED,
        )
        return self.add_node(
            ops.OpType(op),
            None,
            list(type_to_row(instantiation.output)),
            parent,
            [*args, def_port],
        )

    def add_indirect_call(
        self, fun_port: OutPortV, args: list[OutPortV], parent: Node | None = None
    ) -> VNode:
        """Adds an `IndirectCall` node to the graph."""
        assert isinstance(fun_port.ty, FunctionType)

        return self.add_node(
            ops.OpType(ops.CallIndirect(parent=UNDEFINED)),
            None,
            list(type_to_row(fun_port.ty.output)),
            parent,
            [fun_port, *args],
        )

    def add_partial(
        self, def_port: OutPortV, inputs: list[OutPortV], parent: Node | None = None
    ) -> VNode:
        """Adds a `Partial` evaluation node to the graph."""
        assert isinstance(def_port.ty, FunctionType)
        assert len(def_port.ty.inputs) >= len(inputs)
        assert [p.ty.to_hugr() for p in inputs] == [
            ty.to_hugr() for ty in def_port.ty.inputs[: len(inputs)]
        ]
        new_ty = FunctionType(
            def_port.ty.inputs[len(inputs) :],
            def_port.ty.output,
            def_port.ty.input_names[len(inputs) :]
            if def_port.ty.input_names is not None
            else None,
        )
        return self.add_node(
            DummyOp("partial"), None, [new_ty], parent, [*inputs, def_port]
        )

    def add_def(
        self, fun_ty: FunctionType, parent: Node | None, name: str
    ) -> DFContainingVNode:
        """Adds a `FucnDefn` node to the graph."""
        op = ops.FuncDefn(name=name, signature=fun_ty.to_hugr_poly(), parent=UNDEFINED)
        return self._add_dfg_node(ops.OpType(op), [], [fun_ty], parent, None)

    def add_declare(self, fun_ty: FunctionType, parent: Node, name: str) -> VNode:
        """Adds a `FuncDecl` node to the graph."""
        op = ops.FuncDecl(name=name, signature=fun_ty.to_hugr_poly(), parent=UNDEFINED)
        return self.add_node(ops.OpType(op), [], [fun_ty], parent, None)

    def add_edge(self, src_port: OutPort, tgt_port: InPort) -> None:
        """Adds an edge between two ports."""
        if isinstance(src_port, OutPortV) or isinstance(tgt_port, InPortV):
            assert isinstance(src_port, OutPortV)
            assert isinstance(tgt_port, InPortV)
            assert src_port.ty == tgt_port.ty
        else:
            assert isinstance(src_port, OutPortCF)
            assert isinstance(tgt_port, InPortCF)
        self._graph.add_edge(
            src_port.node.idx, tgt_port.node.idx, key=(src_port.offset, tgt_port.offset)
        )

    def add_order_edge(self, src: Node, tgt: Node) -> None:
        """Adds a order-edge between two nodes."""
        self._graph.add_edge(src.idx, tgt.idx, key=ORDER_EDGE_KEY)

    def nodes(self) -> Iterator[Node]:
        """Returns an iterator over all nodes in the graph."""
        return (n["data"] for n in self._graph.nodes.values())

    def get_node(self, idx: int) -> Node:
        """Returns the node corresponding to given index."""
        return self._graph.nodes[idx]["data"]  # type: ignore[no-any-return]

    def children(self, node: Node) -> list[Node]:
        """Returns list of a node's immediate children in the hierarchy."""
        return self._children[node.idx]

    def top_level_nodes(self) -> list[Node]:
        """Returns list of nodes at the top level of the hierarchy.

        These are nodes that do not have a parent. Usually this will just
        be the `Root` node.
        """
        return self._children[-1]

    def edges(self) -> Iterator[Edge]:
        """Returns an iterator over all edges in the graph."""
        return (
            self._to_edge(*e)
            for e in self._graph.edges(keys=True)
            if e[2] != ORDER_EDGE_KEY
        )

    def order_edges(self) -> Iterator[OrderEdge]:
        """Returns an iterator over all order-edges in the graph."""
        return (
            (self.get_node(e[0]), self.get_node(e[1]))
            for e in self._graph.edges(keys=True)
            if e[2] == ORDER_EDGE_KEY
        )

    def in_edges(self, port: InPort) -> Iterator[Edge]:
        """Returns an iterator over all edges connected to a given in-port."""
        for e in self._graph.in_edges(port.node.idx, keys=True):
            if e[2] == ORDER_EDGE_KEY:
                continue
            src, tgt = self._to_edge(*e)
            if tgt.offset == port.offset:
                yield src, tgt

    def out_edges(self, port: OutPort) -> Iterator[Edge]:
        """Returns an iterator over all edges originating from a given out-port."""
        for e in self._graph.out_edges(port.node.idx, keys=True):
            if e[2] == ORDER_EDGE_KEY:
                continue
            src, tgt = self._to_edge(*e)
            if src.offset == port.offset:
                yield src, tgt

    def order_successors(self, node: Node) -> Iterator[Node]:
        """Returns an iterator over all nodes that this node connects to via an
        order edge."""
        for _src, tgt, key in self._graph.out_edges(node.idx, keys=True):
            if key == ORDER_EDGE_KEY:
                yield tgt

    def order_predecessors(self, node: Node) -> Iterator[Node]:
        """Returns an iterator over all nodes that are connected to this node via an
        order edge."""
        for src, _tgt, key in self._graph.in_edges(node.idx, keys=True):
            if key == ORDER_EDGE_KEY:
                yield src

    def _to_edge(self, src: int, tgt: int, key: tuple[int, int]) -> Edge:
        src_node = self.get_node(src)
        tgt_node = self.get_node(tgt)
        return src_node.out_port(key[0]), tgt_node.in_port(key[1])

    def remove_edge(self, src_port: OutPort, tgt_port: InPort) -> None:
        """Removes an edge from the graph."""
        self._graph.remove_edge(
            src_port.node.idx, tgt_port.node.idx, key=(src_port.offset, tgt_port.offset)
        )

    def remove_dummy_nodes(self) -> "Hugr":
        """Replaces dummy ops with external function calls."""
        if self.root is None:
            raise ValueError("Dummy node removal requires a module root node")
        used_names: dict[str, int] = {}
        for n in list(self.nodes()):
            if isinstance(n, VNode) and isinstance(n.op, DummyOp):
                name = n.op.name
                fun_ty = FunctionType(
                    list(n.in_port_types), row_to_type(n.out_port_types)
                )
                if name in used_names:
                    used_names[name] += 1
                    name = f"{name}${used_names[name]}"
                else:
                    used_names[name] = 0
                decl = self.add_declare(fun_ty, self.root, name)
                n.op = ops.OpType(
                    ops.Call(
                        func_sig=fun_ty.to_hugr_poly(),
                        type_args=[],
                        instantiation=fun_ty.to_hugr().root,
                        parent=UNDEFINED,
                    )
                )
                self.add_edge(decl.out_port(0), n.add_in_port(fun_ty))
        return self

    def insert_order_edges(self) -> "Hugr":
        """Adds order edges to the source and target inter-graph edges.

        This ensures that the source is executed before the target. This action must be
        performed before serialisation.
        """
        for src, tgt in list(self.edges()):
            # Exclude CF and constant edges
            if isinstance(src, OutPortCF) or isinstance(
                src.node.op.root, ops.FuncDecl | ops.FuncDefn | ops.Const
            ):
                continue

            if src.node.parent != tgt.node.parent:
                # Walk up the hierarchy from the tgt until we hit a node at the same
                # level as src
                node = tgt.node
                while node.parent != src.node.parent:
                    if node.parent is None:
                        raise ValueError("Invalid non-local edge!")
                    node = node.parent
                # Add order edge to make sure that the src is executed first
                self.add_order_edge(src.node, node)
        return self

    def to_raw(self) -> raw.SerialHugr:
        """Returns the raw representation of this HUGR for serialisation."""
        if self.root is None:
            raise ValueError("Serial Hugr requires a root node")

        self.remove_dummy_nodes()
        self.insert_order_edges()
        # Hugr requires that Input/Output nodes are the first/second children in a DFG.
        # Furthermore, exit nodes must be the second children of CFGs. We're going to
        # satisfy this trivially by first serialising all inputs, outputs, entry and
        # exit nodes
        input_nodes: list[Node] = []
        output_nodes: list[Node] = []
        entry_nodes: list[Node] = []
        exit_nodes: list[Node] = []
        remaining_nodes: list[Node] = []
        indices = itertools.count()
        raw_index: dict[int, node_port.NodeIdx] = {}
        all_nodes = self.nodes()
        for n in all_nodes:
            if n is self.root:
                continue
            match n.op.root:
                case ops.Input():
                    input_nodes.append(n)
                case ops.Output():
                    output_nodes.append(n)
                # We can detect entry BBs by looking for BBs without incoming edges
                # since Guppy will never generate an edge pointing back to the entry.
                # Also, Guppy errors on unreachable code, so we will never generate
                # interior BBs without incoming edges. Hence, there are also no false
                # positives.
                case ops.DataflowBlock() if next(
                    self.in_edges(n.in_port(None)), None
                ) is None:
                    entry_nodes.append(n)
                case ops.ExitBlock():
                    exit_nodes.append(n)
                case _:
                    remaining_nodes.append(n)
        for n in itertools.chain(
            iter([self.root]),
            iter(entry_nodes),
            iter(exit_nodes),
            iter(input_nodes),
            iter(output_nodes),
            iter(remaining_nodes),
        ):
            raw_index[n.idx] = next(indices)

        nodes: list[ops.OpType] = [
            ops.OpType(ops.Module(parent=UNDEFINED))
        ] * self._graph.number_of_nodes()
        for n in self.nodes():
            idx = raw_index[n.idx]
            # Nodes without parent have themselves as parent in the serialised format
            parent = n.parent or n
            n.update_op()
            n.op.root.parent = raw_index[parent.idx]
            nodes[idx] = n.op

        edges: list[raw.Edge] = []
        for src, tgt in self.edges():
            edges.append(
                (
                    (raw_index[src.node.idx], src.offset),
                    (raw_index[tgt.node.idx], tgt.offset),
                )
            )

        for src, tgt in self.order_edges():
            edges.append(((raw_index[src.idx], None), (raw_index[tgt.idx], None)))

        return raw.SerialHugr(nodes=nodes, edges=edges)

    def serialize(self) -> str:
        """Serialize this Hugr in JSON format."""
        return self.to_raw().to_json()



"""Type checking code for statements.

Operates on statements in a basic block after CFG construction. In particular, we
assume that statements involving control flow (i.e. if, while, break, and return
statements) have been removed during CFG construction.

After checking, we return a desugared statement where all sub-expression have been type
annotated.
"""

import ast
from collections.abc import Sequence

from guppylang.ast_util import AstVisitor, with_loc
from guppylang.cfg.bb import BB, BBStatement
from guppylang.checker.core import Context, FieldAccess, Variable
from guppylang.checker.expr_checker import ExprChecker, ExprSynthesizer
from guppylang.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang.nodes import NestedFunctionDef, PlaceNode
from guppylang.tys.parsing import type_from_ast
from guppylang.tys.subst import Subst
from guppylang.tys.ty import NoneType, StructType, TupleType, Type


class StmtChecker(AstVisitor[BBStatement]):
    ctx: Context
    bb: BB | None
    return_ty: Type | None

    def __init__(
        self, ctx: Context, bb: BB | None = None, return_ty: Type | None = None
    ) -> None:
        assert not return_ty or not return_ty.unsolved_vars
        self.ctx = ctx
        self.bb = bb
        self.return_ty = return_ty

    def check_stmts(self, stmts: Sequence[BBStatement]) -> list[BBStatement]:
        return [self.visit(s) for s in stmts]

    def _synth_expr(self, node: ast.expr) -> tuple[ast.expr, Type]:
        return ExprSynthesizer(self.ctx).synthesize(node)

    def _check_expr(
        self, node: ast.expr, ty: Type, kind: str = "expression"
    ) -> tuple[ast.expr, Subst]:
        return ExprChecker(self.ctx).check(node, ty, kind)

    def _check_assign(self, lhs: ast.expr, ty: Type, node: ast.stmt) -> ast.expr:
        """Helper function to check assignments with patterns."""
        match lhs:
            # Easiest case is if the LHS pattern is a single variable.
            case ast.Name(id=x):
                var = Variable(x, ty, lhs)
                self.ctx.locals[x] = var
                return with_loc(lhs, PlaceNode(place=var))

            # The LHS could also be a field `expr.field`
            case ast.Attribute(value=value, attr=attr):
                value, struct_ty = self._synth_expr(value)
                if (
                    not isinstance(struct_ty, StructType)
                    or attr not in struct_ty.field_dict
                ):
                    raise GuppyTypeError(
                        f"Expression of type `{struct_ty}` has no attribute `{attr}`",
                        # Unfortunately, `attr` doesn't contain source annotations, so
                        # we have to use `lhs` as the error location
                        lhs,
                    )
                field = struct_ty.field_dict[attr]
                # TODO: In the future, we could infer some type args here
                if field.ty != ty:
                    raise GuppyTypeError(
                        f"Cannot assign expression of type `{ty}` to field with type "
                        f"`{field.ty}`",
                        lhs,
                    )
                if not isinstance(value, PlaceNode):
                    # For now we complain if someone tries to assign to something that
                    # is not a place, e.g. `f().a = 4`. This would only make sense if
                    # there is another reference to the return value of `f`, otherwise
                    # the mutation cannot be observed. We can start supporting this once
                    # we have proper reference semantics.
                    raise GuppyError(
                        "Assigning to this expression is not supported yet. Consider "
                        "binding the expression to variable and mutate that variable "
                        "instead.",
                        value,
                    )
                if not field.ty.linear:
                    raise GuppyError(
                        "Mutation of classical fields is not supported yet", lhs
                    )
                place = FieldAccess(value.place, struct_ty.field_dict[attr], lhs)
                return with_loc(lhs, PlaceNode(place=place))

            # The only other thing we support right now are tuples
            case ast.Tuple(elts=elts) as lhs:
                tys = ty.element_types if isinstance(ty, TupleType) else [ty]
                n, m = len(elts), len(tys)
                if n != m:
                    raise GuppyTypeError(
                        f"{'Too many' if n < m else 'Not enough'} values to unpack "
                        f"(expected {n}, got {m})",
                        node,
                    )
                lhs.elts = [
                    self._check_assign(pat, el_ty, node)
                    for pat, el_ty in zip(elts, tys, strict=True)
                ]
                return lhs

            # TODO: Python also supports assignments like `[a, b] = [1, 2]` or
            #  `a, *b = ...`. The former would require some runtime checks but
            #  the latter should be easier to do (unpack and repack the rest).
            case _:
                raise GuppyError("Assignment pattern not supported", lhs)

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        if len(node.targets) > 1:
            # This is the case for assignments like `a = b = 1`
            raise GuppyError("Multi assignment not supported", node)

        [target] = node.targets
        node.value, ty = self._synth_expr(node.value)
        node.targets = [self._check_assign(target, ty, node)]
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.stmt:
        if node.value is None:
            raise GuppyError(
                "Variable declaration is not supported. Assignment is required", node
            )
        ty = type_from_ast(node.annotation, self.ctx.globals)
        node.value, subst = self._check_expr(node.value, ty)
        assert not ty.unsolved_vars  # `ty` must be closed!
        assert len(subst) == 0
        target = self._check_assign(node.target, ty, node)
        return with_loc(node, ast.Assign(targets=[target], value=node.value))

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.stmt:
        bin_op = with_loc(
            node, ast.BinOp(left=node.target, op=node.op, right=node.value)
        )
        assign = with_loc(node, ast.Assign(targets=[node.target], value=bin_op))
        return self.visit_Assign(assign)

    def visit_Expr(self, node: ast.Expr) -> ast.stmt:
        # An expression statement where the return value is discarded
        node.value, ty = self._synth_expr(node.value)
        if ty.linear:
            raise GuppyTypeError(f"Value with linear type `{ty}` is not used", node)
        return node

    def visit_Return(self, node: ast.Return) -> ast.stmt:
        if not self.return_ty:
            raise InternalGuppyError("return_ty required to check return stmt!")

        if node.value is not None:
            node.value, subst = self._check_expr(
                node.value, self.return_ty, "return value"
            )
            assert len(subst) == 0  # `self.return_ty` is closed!
        elif not isinstance(self.return_ty, NoneType):
            raise GuppyTypeError(
                f"Expected return value of type `{self.return_ty}`", None
            )
        return node

    def visit_NestedFunctionDef(self, node: NestedFunctionDef) -> ast.stmt:
        from guppylang.checker.func_checker import check_nested_func_def

        if not self.bb:
            raise InternalGuppyError("BB required to check nested function def!")

        func_def = check_nested_func_def(node, self.bb, self.ctx)
        self.ctx.locals[func_def.name] = Variable(func_def.name, func_def.ty, func_def)
        return func_def

    def visit_If(self, node: ast.If) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_While(self, node: ast.While) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_Break(self, node: ast.Break) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_Continue(self, node: ast.Continue) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")



import ast
from collections.abc import Sequence

from guppylang.ast_util import AstVisitor, get_type
from guppylang.checker.core import Variable
from guppylang.compiler.core import (
    CompiledGlobals,
    CompilerBase,
    DFContainer,
    return_var,
)
from guppylang.compiler.expr_compiler import ExprCompiler
from guppylang.error import InternalGuppyError
from guppylang.hugr_builder.hugr import Hugr, OutPortV
from guppylang.nodes import CheckedNestedFunctionDef, PlaceNode
from guppylang.tys.ty import TupleType


class StmtCompiler(CompilerBase, AstVisitor[None]):
    """A compiler for Guppy statements to Hugr"""

    expr_compiler: ExprCompiler

    dfg: DFContainer

    def __init__(self, graph: Hugr, globals: CompiledGlobals):
        super().__init__(graph, globals)
        self.expr_compiler = ExprCompiler(graph, globals)

    def compile_stmts(
        self,
        stmts: Sequence[ast.stmt],
        dfg: DFContainer,
    ) -> DFContainer:
        """Compiles a list of basic statements into a dataflow node.

        Note that the `dfg` is mutated in-place. After compilation, the DFG will also
        contain all variables that are assigned in the given list of statements.
        """
        self.dfg = dfg
        for s in stmts:
            self.visit(s)
        return self.dfg

    def _unpack_assign(self, lhs: ast.expr, port: OutPortV, node: ast.stmt) -> None:
        """Updates the local DFG with assignments."""
        if isinstance(lhs, PlaceNode):
            self.dfg[lhs.place] = port
        elif isinstance(lhs, ast.Tuple):
            unpack = self.graph.add_unpack_tuple(port, self.dfg.node)
            for i, pat in enumerate(lhs.elts):
                self._unpack_assign(pat, unpack.out_port(i), node)
        else:
            raise InternalGuppyError("Invalid assign pattern in compiler")

    def visit_Assign(self, node: ast.Assign) -> None:
        [target] = node.targets
        port = self.expr_compiler.compile(node.value, self.dfg)
        self._unpack_assign(target, port, node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        assert node.value is not None
        port = self.expr_compiler.compile(node.value, self.dfg)
        self._unpack_assign(node.target, port, node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Expr(self, node: ast.Expr) -> None:
        self.expr_compiler.compile_row(node.value, self.dfg)

    def visit_Return(self, node: ast.Return) -> None:
        # We turn returns into assignments of dummy variables, i.e. the statement
        # `return e0, e1, e2` is turned into `%ret0 = e0; %ret1 = e1; %ret2 = e2`.
        if node.value is not None:
            return_ty = get_type(node.value)
            port = self.expr_compiler.compile(node.value, self.dfg)
            if isinstance(return_ty, TupleType):
                unpack = self.graph.add_unpack_tuple(port, self.dfg.node)
                row = [unpack.out_port(i) for i in range(len(return_ty.element_types))]
            else:
                row = [port]
            for i, port in enumerate(row):
                var = Variable(return_var(i), port.ty, node.value)
                self.dfg[var] = port

    def visit_CheckedNestedFunctionDef(self, node: CheckedNestedFunctionDef) -> None:
        from guppylang.compiler.func_compiler import compile_local_func_def

        var = Variable(node.name, node.ty, node)
        self.dfg[var] = compile_local_func_def(node, self.dfg, self.graph, self.globals)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, ClassVar, TypeAlias, cast

from hugr.serialization import tys
from hugr.serialization.tys import TypeBound

from guppylang.error import InternalGuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.common import ToHugr, Transformable, Transformer, Visitor
from guppylang.tys.const import Const, ConstValue, ExistentialConstVar
from guppylang.tys.param import Parameter
from guppylang.tys.var import BoundVar, ExistentialVar

if TYPE_CHECKING:
    from guppylang.definition.struct import CheckedStructDef, StructField
    from guppylang.definition.ty import OpaqueTypeDef
    from guppylang.tys.subst import Inst, Subst


@dataclass(frozen=True)
class TypeBase(ToHugr[tys.Type], Transformable["Type"], ABC):
    """Abstract base class for all Guppy types.

    Note that all subclasses are expected to be immutable.
    """

    @cached_property
    @abstractmethod
    def linear(self) -> bool:
        """Whether this type should be treated linearly."""

    @cached_property
    @abstractmethod
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`.

        This needs to be specified explicitly, since opaque nonlinear types in a Hugr
        extension could be either declared as copyable or equatable. If we don't get the
        bound exactly right during serialisation, the Hugr validator will complain.
        """

    @abstractmethod
    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`.

        This enforces that all implementors of `TypeBase` can be embedded into the
        `Type` union type.
        """

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return set()

    def substitute(self, subst: "Subst") -> "Type":
        """Substitutes existential variables in this type."""
        from guppylang.tys.subst import Substituter

        return self.transform(Substituter(subst))

    def to_arg(self) -> TypeArg:
        """Wraps this constant into a type argument."""
        return TypeArg(self.cast())

    def __str__(self) -> str:
        """Returns a human-readable representation of the type."""
        from guppylang.tys.printing import TypePrinter

        # We use a custom printer that takes care of inserting parentheses and choosing
        # unique names
        return TypePrinter().visit(cast(Type, self))


@dataclass(frozen=True)
class ParametrizedTypeBase(TypeBase, ABC):
    """Abstract base class for types that depend on parameters.

    For example, `list`, `tuple`, etc. require arguments in order to be turned into a
    proper type.

    Note that all subclasses are expected to be immutable.
    """

    args: Sequence[Argument]

    def __post_init__(self) -> None:
        # Make sure that we don't have nested generic functions
        for arg in self.args:
            match arg:
                case TypeArg(ty=FunctionType(parametrized=True)):
                    raise InternalGuppyError(
                        "Tried to construct a higher-rank polymorphic type!"
                    )

    @property
    @abstractmethod
    def intrinsically_linear(self) -> bool:
        """Whether this type is linear, independent of the arguments.

        For example, a parametrized struct containing a qubit is linear, no matter what
        the arguments are.
        """

    @cached_property
    def linear(self) -> bool:
        """Whether this type should be treated linearly."""
        return self.intrinsically_linear or any(
            isinstance(arg, TypeArg) and arg.ty.linear for arg in self.args
        )

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return set().union(*(arg.unsolved_vars for arg in self.args))

    @cached_property
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.linear:
            return tys.TypeBound.Any
        return tys.TypeBound.join(
            *(arg.ty.hugr_bound for arg in self.args if isinstance(arg, TypeArg))
        )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        if not visitor.visit(self):
            for arg in self.args:
                visitor.visit(arg)


@dataclass(frozen=True)
class BoundTypeVar(TypeBase, BoundVar):
    """Bound type variable, referencing a parameter of kind `Type`.

    For example, in the function type `forall T. list[T] -> T` we represent `T` as a
    `BoundTypeVar(idx=0)`.

    A bound type variables can be instantiated with a `TypeArg` argument.
    """

    linear: bool

    @cached_property
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.linear:
            return TypeBound.Any
        # We're conservative and don't require equatability for non-linear variables.
        # This is fine since Guppy doesn't use the equatable feature anyways.
        return TypeBound.Copyable

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        return tys.Type(tys.Variable(i=self.idx, b=self.hugr_bound))

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self

    def __str__(self) -> str:
        """Returns a human-readable representation of the type."""
        return self.display_name


@dataclass(frozen=True)
class ExistentialTypeVar(ExistentialVar, TypeBase):
    """Existential type variable.

    For example, the empty list literal `[]` is typed as `list[?T]` where `?T` stands
    for an existential type variable.

    During type checking we try to solve all existential type variables and substitute
    them with concrete types.
    """

    linear: bool

    @classmethod
    def fresh(cls, display_name: str, linear: bool) -> "ExistentialTypeVar":
        return ExistentialTypeVar(display_name, next(cls._fresh_id), linear)

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return {self}

    @cached_property
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        raise InternalGuppyError(
            "Tried to compute bound of unsolved existential type variable"
        )

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        raise InternalGuppyError(
            "Tried to convert unsolved existential type variable to Hugr"
        )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


@dataclass(frozen=True)
class NoneType(TypeBase):
    """Type of tuples."""

    linear: bool = field(default=False, init=False)
    hugr_bound: tys.TypeBound = field(default=tys.TypeBound.Copyable, init=False)

    # Flag to avoid turning the type into a row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to Nones are not broken up into
    # empty rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        return TupleType([]).to_hugr()

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


@dataclass(frozen=True)
class NumericType(TypeBase):
    """Numeric types like `int` and `float`."""

    kind: "Kind"

    class Kind(Enum):
        """The different kinds of numeric types."""

        Nat = "nat"
        Int = "int"
        Float = "float"

    INT_WIDTH: ClassVar[int] = 6

    @property
    def linear(self) -> bool:
        """Whether this type should be treated linearly."""
        return False

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        match self.kind:
            case NumericType.Kind.Nat | NumericType.Kind.Int:
                return tys.Type(
                    tys.Opaque(
                        extension="arithmetic.int.types",
                        id="int",
                        args=[tys.TypeArg(tys.BoundedNatArg(n=NumericType.INT_WIDTH))],
                        bound=tys.TypeBound.Copyable,
                    )
                )
            case NumericType.Kind.Float:
                return tys.Type(
                    tys.Opaque(
                        extension="arithmetic.float.types",
                        id="float64",
                        args=[],
                        bound=tys.TypeBound.Copyable,
                    )
                )

    @property
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any` or `Copyable`"""
        return tys.TypeBound.Copyable

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


@dataclass(frozen=True, init=False)
class FunctionType(ParametrizedTypeBase):
    """Type of (potentially generic) functions."""

    inputs: Sequence["Type"]
    output: "Type"
    params: Sequence[Parameter]
    input_names: Sequence[str] | None

    args: Sequence[Argument] = field(init=False)
    linear: bool = field(default=False, init=False)
    intrinsically_linear: bool = field(default=False, init=False)
    hugr_bound: tys.TypeBound = field(default=TypeBound.Copyable, init=False)

    def __init__(
        self,
        inputs: Sequence["Type"],
        output: "Type",
        input_names: Sequence[str] | None = None,
        params: Sequence[Parameter] | None = None,
    ) -> None:
        # We need a custom __init__ to set the args
        args = [TypeArg(ty) for ty in inputs]
        args.append(TypeArg(output))
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "inputs", inputs)
        object.__setattr__(self, "output", output)
        object.__setattr__(self, "input_names", input_names or [])
        object.__setattr__(self, "params", params or [])

    @property
    def parametrized(self) -> bool:
        """Whether the function is parametrized."""
        return len(self.params) > 0

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        if self.parametrized:
            raise InternalGuppyError(
                "Tried to convert parametrised function type to Hugr. Use "
                "`to_hugr_poly` instead"
            )
        return tys.Type(self._to_hugr_function_type())

    def to_hugr_poly(self) -> tys.PolyFuncType:
        """Computes the Hugr `PolyFuncType` representation of the type."""
        func_ty = self._to_hugr_function_type()
        return tys.PolyFuncType(params=[p.to_hugr() for p in self.params], body=func_ty)

    def _to_hugr_function_type(self) -> tys.FunctionType:
        """Helper method to compute the Hugr `FunctionType` representation of the type.

        The resulting `FunctionType` can then be embedded into a Hugr `Type` or a Hugr
        `PolyFuncType`.
        """
        ins = [t.to_hugr() for t in self.inputs]
        outs = [t.to_hugr() for t in type_to_row(self.output)]
        return tys.FunctionType(input=ins, output=outs)

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        if not visitor.visit(self):
            for inp in self.inputs:
                visitor.visit(inp)
            visitor.visit(self.output)
            for param in self.params:
                visitor.visit(param)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or FunctionType(
            [inp.transform(transformer) for inp in self.inputs],
            self.output.transform(transformer),
            self.input_names,
            self.params,
        )

    def instantiate(self, args: "Inst") -> "FunctionType":
        """Instantiates all function parameters with concrete types."""
        from guppylang.tys.subst import Instantiator

        assert len(args) == len(self.params)

        # Set the `preserve` flag for instantiated tuples and None
        preserved_args: list[Argument] = []
        for arg in args:
            if isinstance(arg, TypeArg):
                if isinstance(arg.ty, TupleType):
                    arg = TypeArg(TupleType(arg.ty.element_types, preserve=True))
                elif isinstance(arg.ty, NoneType):
                    arg = TypeArg(NoneType(preserve=True))
            preserved_args.append(arg)

        inst = Instantiator(preserved_args)
        return FunctionType(
            [ty.transform(inst) for ty in self.inputs],
            self.output.transform(inst),
            self.input_names,
        )

    def unquantified(self) -> tuple["FunctionType", Sequence[ExistentialVar]]:
        """Instantiates all parameters with existential variables."""
        exs = [param.to_existential() for param in self.params]
        return self.instantiate([arg for arg, _ in exs]), [var for _, var in exs]


@dataclass(frozen=True, init=False)
class TupleType(ParametrizedTypeBase):
    """Type of tuples."""

    element_types: Sequence["Type"]

    # Flag to avoid turning the tuple into a row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to tuples are not broken up into
    # rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    def __init__(self, element_types: Sequence["Type"], preserve: bool = False) -> None:
        # We need a custom __init__ to set the args
        args = [TypeArg(ty) for ty in element_types]
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "element_types", element_types)
        object.__setattr__(self, "preserve", preserve)

    @property
    def intrinsically_linear(self) -> bool:
        return False

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        # Tuples are encoded as a unary sum. Note that we need to make a copy of this
        # tuple with `preserve=False` to ensure that it can be broken up into a row (if
        # this tuple was created by instantiating a type variable, it is still
        # represented as a *row* sum).
        tuple_ty = TupleType(self.element_types, preserve=False)
        return SumType([tuple_ty]).to_hugr()

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or TupleType(
            [ty.transform(transformer) for ty in self.element_types], self.preserve
        )


@dataclass(frozen=True, init=False)
class SumType(ParametrizedTypeBase):
    """Type of sums.

    Note that this type is only used internally when constructing the Hugr. Users cannot
    write down this type.
    """

    element_types: Sequence["Type"]

    def __init__(self, element_types: Sequence["Type"]) -> None:
        # We need a custom __init__ to set the args
        args = [TypeArg(ty) for ty in element_types]
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "element_types", element_types)

    @property
    def intrinsically_linear(self) -> bool:
        return False

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        rows = [type_to_row(ty) for ty in self.element_types]
        sum_inner: tys.UnitSum | tys.GeneralSum
        if all(len(row) == 0 for row in rows):
            sum_inner = tys.UnitSum(size=len(rows))
        else:
            sum_inner = tys.GeneralSum(rows=rows_to_hugr(rows))
        return tys.Type(tys.SumType(sum_inner))

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or SumType(
            [ty.transform(transformer) for ty in self.element_types]
        )


@dataclass(frozen=True)
class OpaqueType(ParametrizedTypeBase):
    """Type that is directly backed by a Hugr opaque type.

    For example, many builtin types like `int`, `float`, `list` etc. are directly backed
    by a Hugr extension.
    """

    defn: "OpaqueTypeDef"

    @property
    def intrinsically_linear(self) -> bool:
        """Whether this type is linear, independent of the arguments."""
        return self.defn.always_linear

    @property
    def hugr_bound(self) -> tys.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.defn.bound is not None:
            return self.defn.bound
        return super().hugr_bound

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        return self.defn.to_hugr(self.args)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or OpaqueType(
            [arg.transform(transformer) for arg in self.args], self.defn
        )


@dataclass(frozen=True)
class StructType(ParametrizedTypeBase):
    """A struct type."""

    defn: "CheckedStructDef"

    @cached_property
    def fields(self) -> list["StructField"]:
        """The fields of this struct type."""
        from guppylang.definition.struct import StructField
        from guppylang.tys.subst import Instantiator

        inst = Instantiator(self.args)
        return [StructField(f.name, f.ty.transform(inst)) for f in self.defn.fields]

    @cached_property
    def field_dict(self) -> "dict[str, StructField]":
        """Mapping from names to fields of this struct type."""
        return {field.name: field for field in self.fields}

    @cached_property
    def intrinsically_linear(self) -> bool:
        """Whether this type is linear, independent of the arguments."""
        return any(f.ty.linear for f in self.defn.fields)

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self) -> tys.Type:
        """Computes the Hugr representation of the type."""
        return TupleType([f.ty for f in self.fields]).to_hugr()

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or StructType(
            [arg.transform(transformer) for arg in self.args], self.defn
        )


#: The type of parametrized Guppy types.
ParametrizedType: TypeAlias = (
    FunctionType | TupleType | SumType | OpaqueType | StructType
)

#: The type of Guppy types.
#:
#: This is a type alias for a union of all Guppy types defined in this module. This
#: models an algebraic data type and enables exhaustiveness checking in pattern matches
#: etc.
#:
#: This might become obsolete in case the @sealed decorator is added:
#:   * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#:   * https://github.com/johnthagen/sealed-typing-pep
Type: TypeAlias = (
    BoundTypeVar | ExistentialTypeVar | NumericType | NoneType | ParametrizedType
)

#: An immutable row of Guppy types.
TypeRow: TypeAlias = Sequence[Type]


def row_to_type(row: TypeRow) -> Type:
    """Turns a row of types into a single type by packing into a tuple."""
    if len(row) == 0:
        return NoneType()
    elif len(row) == 1:
        return row[0]
    else:
        return TupleType(row)


def type_to_row(ty: Type) -> TypeRow:
    """Turns a type into a row of types by unpacking top-level tuples."""
    if isinstance(ty, NoneType) and not ty.preserve:
        return []
    if isinstance(ty, TupleType) and not ty.preserve:
        return ty.element_types
    return [ty]


def row_to_hugr(row: TypeRow) -> tys.TypeRow:
    """Computes the Hugr representation of a type row."""
    return [ty.to_hugr() for ty in row]


def rows_to_hugr(rows: Sequence[TypeRow]) -> list[tys.TypeRow]:
    """Computes the Hugr representation of a sequence of rows."""
    return [row_to_hugr(row) for row in rows]


def unify(s: Type | Const, t: Type | Const, subst: "Subst | None") -> "Subst | None":
    """Computes a most general unifier for two types or constants.

    Return a substitutions `subst` such that `s[subst] == t[subst]` or `None` if this
    not possible.
    """
    # Make sure that s and t are either both constants or both types
    assert isinstance(s, TypeBase) == isinstance(t, TypeBase)
    if subst is None:
        return None
    match s, t:
        case ExistentialVar(id=s_id), ExistentialVar(id=t_id) if s_id == t_id:
            return subst
        case ExistentialTypeVar() | ExistentialConstVar() as s_var, t:
            return _unify_var(s_var, t, subst)
        case s, ExistentialTypeVar() | ExistentialConstVar() as t_var:
            return _unify_var(t_var, s, subst)
        case BoundVar(idx=s_idx), BoundVar(idx=t_idx) if s_idx == t_idx:
            return subst
        case ConstValue(value=c_value), ConstValue(value=d_value) if c_value == d_value:
            return subst
        case NumericType(kind=s_kind), NumericType(kind=t_kind) if s_kind == t_kind:
            return subst
        case NoneType(), NoneType():
            return subst
        case FunctionType() as s, FunctionType() as t if s.params == t.params:
            return _unify_args(s, t, subst)
        case TupleType() as s, TupleType() as t:
            return _unify_args(s, t, subst)
        case SumType() as s, SumType() as t:
            return _unify_args(s, t, subst)
        case OpaqueType() as s, OpaqueType() as t if s.defn == t.defn:
            return _unify_args(s, t, subst)
        case StructType() as s, StructType() as t if s.defn == t.defn:
            return _unify_args(s, t, subst)
        case _:
            return None


def _unify_var(
    var: ExistentialTypeVar | ExistentialConstVar, t: Type | Const, subst: "Subst"
) -> "Subst | None":
    """Helper function for unification of type or const variables."""
    if var in subst:
        return unify(subst[var], t, subst)
    if isinstance(t, ExistentialTypeVar) and t in subst:
        return unify(var, subst[t], subst)
    if var in t.unsolved_vars:
        return None
    return {var: t, **subst}


def _unify_args(
    s: ParametrizedType, t: ParametrizedType, subst: "Subst"
) -> "Subst | None":
    """Helper function for unification of type arguments of parametrised types."""
    if len(s.args) != len(t.args):
        return None
    for sa, ta in zip(s.args, t.args, strict=True):
        match sa, ta:
            case TypeArg(ty=sa_ty), TypeArg(ty=ta_ty):
                res = unify(sa_ty, ta_ty, subst)
                if res is None:
                    return None
                subst = res
            case ConstArg(const=sa_const), ConstArg(const=ta_const):
                res = unify(sa_const, ta_const, subst)
                if res is None:
                    return None
                subst = res
            case _:
                return None
    return subst


### Helpers for working with tuples of functions


def parse_function_tensor(ty: TupleType) -> list[FunctionType] | None:
    """Parses a nested tuple of function types into a flat list of functions."""
    result = []
    for el in ty.element_types:
        if isinstance(el, FunctionType):
            result.append(el)
        elif isinstance(el, TupleType):
            funcs = parse_function_tensor(el)
            if funcs:
                result.extend(funcs)
            else:
                return None
    return result


def function_tensor_signature(tys: list[FunctionType]) -> FunctionType:
    """Compute the combined function signature of a list of functions"""
    inputs: list[Type] = []
    outputs: list[Type] = []
    for fun_ty in tys:
        assert not fun_ty.parametrized
        inputs.extend(fun_ty.inputs)
        outputs.extend(type_to_row(fun_ty.output))
    return FunctionType(inputs, row_to_type(outputs))



import inspect
import itertools
import sys
from collections.abc import Callable, Mapping
from types import ModuleType
from typing import Any, Union

from guppylang.checker.core import Globals, PyScope
from guppylang.compiler.core import CompiledGlobals
from guppylang.definition.common import (
    CheckableDef,
    CheckedDef,
    CompilableDef,
    DefId,
    ParsableDef,
    RawDef,
)
from guppylang.definition.custom import CustomFunctionDef
from guppylang.definition.declaration import RawFunctionDecl
from guppylang.definition.function import RawFunctionDef
from guppylang.definition.parameter import ParamDef
from guppylang.definition.struct import CheckedStructDef
from guppylang.definition.ty import TypeDef
from guppylang.error import GuppyError, pretty_errors
from guppylang.hugr_builder.hugr import Hugr

PyFunc = Callable[..., Any]
PyFuncDefOrDecl = tuple[bool, PyFunc]


class GuppyModule:
    """A Guppy module that may contain function and type definitions."""

    name: str

    # Whether the module has already been compiled
    _compiled: bool

    # If the hugr has already been compiled, keeps a reference that can be returned
    # from `compile`.
    _compiled_hugr: Hugr | None

    # Map of raw definitions in this module
    _raw_defs: dict[DefId, RawDef]
    _raw_type_defs: dict[DefId, RawDef]

    # Globals from imported modules
    _imported_globals: Globals
    _imported_compiled_globals: CompiledGlobals

    # Globals for functions and types defined in this module. Only gets populated during
    # compilation
    _globals: Globals
    _compiled_globals: CompiledGlobals

    # When `_instance_buffer` is not `None`, then all registered functions will be
    # buffered in this list. They only get properly registered, once
    # `_register_buffered_instance_funcs` is called. This way, we can associate
    _instance_func_buffer: dict[str, RawDef] | None

    def __init__(self, name: str, import_builtins: bool = True):
        self.name = name
        self._globals = Globals({}, {}, {}, {})
        self._compiled_globals = {}
        self._imported_globals = Globals.default()
        self._imported_compiled_globals = {}
        self._compiled = False
        self._compiled_hugr = None
        self._instance_func_buffer = None
        self._raw_defs = {}
        self._raw_type_defs = {}

        # Import builtin module
        if import_builtins:
            import guppylang.prelude.builtins as builtins

            self.load(builtins)

    def load(self, m: Union[ModuleType, "GuppyModule"]) -> None:
        """Imports another Guppy module."""
        self._check_not_yet_compiled()
        if isinstance(m, GuppyModule):
            # Compile module if it isn't compiled yet
            if not m.compiled:
                m.compile()

            # For now, we can only import custom functions
            if any(
                not isinstance(v, CustomFunctionDef | TypeDef | ParamDef)
                for v in m._compiled_globals.values()
            ):
                raise GuppyError(
                    "Importing modules with defined functions is not supported yet"
                )

            self._imported_globals |= m._globals
            self._imported_compiled_globals |= m._compiled_globals
        else:
            for val in m.__dict__.values():
                if isinstance(val, GuppyModule):
                    self.load(val)

    def register_def(self, defn: RawDef, instance: TypeDef | None = None) -> None:
        """Registers a definition with this module.

        If the name of the definition is already defined, the new definition
        replaces the old.

        Optionally, the definition can be marked as an instance method by passing the
        corresponding instance type definition.
        """
        self._check_not_yet_compiled()
        if self._instance_func_buffer is not None and not isinstance(defn, TypeDef):
            self._instance_func_buffer[defn.name] = defn
        else:
            if isinstance(defn, TypeDef | ParamDef):
                self._raw_type_defs[defn.id] = defn
            else:
                self._raw_defs[defn.id] = defn
            if instance is not None:
                self._globals.impls.setdefault(instance.id, {})
                self._globals.impls[instance.id][defn.name] = defn.id
            else:
                self._globals.names[defn.name] = defn.id

    def register_func_def(
        self, f: PyFunc, instance: TypeDef | None = None
    ) -> RawFunctionDef:
        """Registers a Python function definition as belonging to this Guppy module."""
        defn = RawFunctionDef(DefId.fresh(self), f.__name__, None, f, get_py_scope(f))
        self.register_def(defn, instance)
        return defn

    def register_func_decl(
        self, f: PyFunc, instance: TypeDef | None = None
    ) -> RawFunctionDecl:
        """Registers a Python function declaration as belonging to this Guppy module."""
        decl = RawFunctionDecl(DefId.fresh(self), f.__name__, None, f)
        self.register_def(decl, instance)
        return decl

    def _register_buffered_instance_funcs(self, instance: TypeDef) -> None:
        assert self._instance_func_buffer is not None
        buffer = self._instance_func_buffer
        self._instance_func_buffer = None
        for defn in buffer.values():
            self.register_def(defn, instance)

    @property
    def compiled(self) -> bool:
        return self._compiled

    @staticmethod
    def _check_defs(
        raw_defs: Mapping[DefId, RawDef], globals: Globals
    ) -> dict[DefId, CheckedDef]:
        """Helper method to parse and check raw definitions."""
        raw_globals = globals | Globals(raw_defs, {}, {}, {})
        parsed = {
            def_id: defn.parse(raw_globals) if isinstance(defn, ParsableDef) else defn
            for def_id, defn in raw_defs.items()
        }
        parsed_globals = globals | Globals(parsed, {}, {}, {})
        return {
            def_id: (
                defn.check(parsed_globals) if isinstance(defn, CheckableDef) else defn
            )
            for def_id, defn in parsed.items()
        }

    @pretty_errors
    def compile(self) -> Hugr:
        """Compiles the module and returns the final Hugr."""
        if self.compiled:
            assert self._compiled_hugr is not None, "Module is compiled but has no Hugr"
            return self._compiled_hugr

        # Type definitions need to be checked first so that we can use them when parsing
        # function signatures etc.
        type_defs = self._check_defs(
            self._raw_type_defs, self._imported_globals | self._globals
        )
        self._globals = self._globals.update_defs(type_defs)

        # Collect auto-generated methods
        generated: dict[DefId, RawDef] = {}
        for defn in type_defs.values():
            if isinstance(defn, CheckedStructDef):
                self._globals.impls.setdefault(defn.id, {})
                for method_def in defn.generated_methods:
                    generated[method_def.id] = method_def
                    self._globals.impls[defn.id][method_def.name] = method_def.id

        # Now, we can check all other definitions
        other_defs = self._check_defs(
            self._raw_defs | generated, self._imported_globals | self._globals
        )
        self._globals = self._globals.update_defs(other_defs)

        # Prepare Hugr for this module
        graph = Hugr(self.name)
        module_node = graph.set_root_name(self.name)

        # Compile definitions to Hugr
        self._compiled_globals = {
            defn.id: (
                defn.compile_outer(graph, module_node)
                if isinstance(defn, CompilableDef)
                else defn
            )
            for defn in itertools.chain(type_defs.values(), other_defs.values())
        }
        all_compiled_globals = self._compiled_globals | self._imported_compiled_globals

        # Finally, compile the definition contents to Hugr. For example, this compiles
        # the bodies of functions.
        for defn in self._compiled_globals.values():
            defn.compile_inner(graph, all_compiled_globals)

        self._compiled = True
        self._compiled_hugr = graph
        return graph

    def contains(self, name: str) -> bool:
        """Returns 'True' if the module contains an object with the given name."""
        return name in self._globals.names

    def _check_not_yet_compiled(self) -> None:
        if self._compiled:
            raise GuppyError(f"The module `{self.name}` has already been compiled")


def get_py_scope(f: PyFunc) -> PyScope:
    """Returns a mapping of all variables captured by a Python function.

    Note that this function only works in CPython. On other platforms, an empty
    dictionary is returned.

    Relies on inspecting the `__globals__` and `__closure__` attributes of the function.
    See https://docs.python.org/3/reference/datamodel.html#special-read-only-attributes
    """
    if sys.implementation.name != "cpython":
        return {}

    if inspect.ismethod(f):
        f = f.__func__
    code = f.__code__

    nonlocals: PyScope = {}
    if f.__closure__ is not None:
        for var, cell in zip(code.co_freevars, f.__closure__, strict=True):
            try:
                value = cell.cell_contents
            except ValueError:
                # The call to `cell_contents` will fail if `var` is a recursive
                # reference to the decorated function
                continue
            nonlocals[var] = value

    return nonlocals | f.__globals__.copy()



"""Type checking code for control-flow graphs

Operates on CFGs produced by the `CFGBuilder`. Produces a `CheckedCFG` consisting of
`CheckedBB`s with inferred type signatures.
"""

import collections
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import Generic, TypeVar

from guppylang.ast_util import line_col
from guppylang.cfg.bb import BB
from guppylang.cfg.cfg import CFG, BaseCFG
from guppylang.checker.core import Context, Globals, Locals, Place, V, Variable
from guppylang.checker.expr_checker import ExprSynthesizer, to_bool
from guppylang.checker.stmt_checker import StmtChecker
from guppylang.error import GuppyError
from guppylang.tys.ty import Type

Row = Sequence[V]


@dataclass(frozen=True)
class Signature(Generic[V]):
    """The signature of a basic block.

    Stores the input/output variables with their types. Generic over the representation
    of program variables.
    """

    input_row: Row[V]
    output_rows: Sequence[Row[V]]  # One for each successor

    @staticmethod
    def empty() -> "Signature[V]":
        return Signature([], [])


@dataclass(eq=False)  # Disable equality to recover hash from `object`
class CheckedBB(BB, Generic[V]):
    """Basic block annotated with an input and output type signature.

    The signature is generic over the representation of program variables.
    """

    sig: Signature[V] = field(default_factory=Signature.empty)


class CheckedCFG(BaseCFG[CheckedBB[V]], Generic[V]):
    input_tys: list[Type]
    output_ty: Type

    def __init__(self, input_tys: list[Type], output_ty: Type) -> None:
        super().__init__([])
        self.input_tys = input_tys
        self.output_ty = output_ty


def check_cfg(
    cfg: CFG, inputs: Row[Variable], return_ty: Type, globals: Globals
) -> CheckedCFG[Place]:
    """Type checks a control-flow graph.

    Annotates the basic blocks with input and output type signatures and removes
    unreachable blocks. Note that the inputs/outputs are annotated in the form of
    *places* rather than just variables.
    """
    # First, we need to run program analysis
    ass_before = {v.name for v in inputs}
    cfg.analyze(ass_before, ass_before)

    # We start by compiling the entry BB
    checked_cfg: CheckedCFG[Variable] = CheckedCFG([v.ty for v in inputs], return_ty)
    checked_cfg.entry_bb = check_bb(
        cfg.entry_bb, checked_cfg, inputs, return_ty, globals
    )
    compiled = {cfg.entry_bb: checked_cfg.entry_bb}

    # Visit all control-flow edges in BFS order. We can't just do a normal loop over
    # all BBs since the input types for a BB are computed by checking a predecessor.
    # We do BFS instead of DFS to get a better error ordering.
    queue = collections.deque(
        (checked_cfg.entry_bb, i, succ)
        # We enumerate the successor starting from the back, so we start with the `True`
        # branch. This way, we find errors in a more natural order
        for i, succ in reverse_enumerate(cfg.entry_bb.successors)
    )
    while len(queue) > 0:
        pred, num_output, bb = queue.popleft()
        input_row = [
            Variable(v.name, v.ty, v.defined_at)
            for v in pred.sig.output_rows[num_output]
        ]

        if bb in compiled:
            # If the BB was already compiled, we just have to check that the signatures
            # match.
            check_rows_match(input_row, compiled[bb].sig.input_row, bb)
        else:
            # Otherwise, check the BB and enqueue its successors
            checked_bb = check_bb(bb, checked_cfg, input_row, return_ty, globals)
            queue += [
                # We enumerate the successor starting from the back, so we start with
                # the `True` branch. This way, we find errors in a more natural order
                (checked_bb, i, succ)
                for i, succ in reverse_enumerate(bb.successors)
            ]
            compiled[bb] = checked_bb

        # Link up BBs in the checked CFG
        compiled[bb].predecessors.append(pred)
        pred.successors[num_output] = compiled[bb]

    checked_cfg.bbs = list(compiled.values())
    checked_cfg.exit_bb = compiled[cfg.exit_bb]  # TODO: Fails if exit is unreachable
    checked_cfg.live_before = {compiled[bb]: cfg.live_before[bb] for bb in cfg.bbs}
    checked_cfg.ass_before = {compiled[bb]: cfg.ass_before[bb] for bb in cfg.bbs}
    checked_cfg.maybe_ass_before = {
        compiled[bb]: cfg.maybe_ass_before[bb] for bb in cfg.bbs
    }

    # Finally, run the linearity check
    from guppylang.checker.linearity_checker import check_cfg_linearity

    linearity_checked_cfg = check_cfg_linearity(checked_cfg)

    return linearity_checked_cfg


def check_bb(
    bb: BB,
    checked_cfg: CheckedCFG[Variable],
    inputs: Row[Variable],
    return_ty: Type,
    globals: Globals,
) -> CheckedBB[Variable]:
    cfg = bb.containing_cfg

    # For the entry BB we have to separately check that all used variables are
    # defined. For all other BBs, this will be checked when compiling a predecessor.
    if bb == cfg.entry_bb:
        assert len(bb.predecessors) == 0
        for x, use in bb.vars.used.items():
            if x not in cfg.ass_before[bb] and x not in globals:
                raise GuppyError(f"Variable `{x}` is not defined", use)

    # Check the basic block
    ctx = Context(globals, Locals({v.name: v for v in inputs}))
    checked_stmts = StmtChecker(ctx, bb, return_ty).check_stmts(bb.statements)

    # If we branch, we also have to check the branch predicate
    if len(bb.successors) > 1:
        assert bb.branch_pred is not None
        bb.branch_pred, ty = ExprSynthesizer(ctx).synthesize(bb.branch_pred)
        bb.branch_pred, _ = to_bool(bb.branch_pred, ty, ctx)

    for succ in bb.successors:
        for x, use_bb in cfg.live_before[succ].items():
            # Check that the variables requested by the successor are defined
            if x not in ctx.locals and x not in ctx.globals:
                # If the variable is defined on *some* paths, we can give a more
                # informative error message
                if x in cfg.maybe_ass_before[use_bb]:
                    # TODO: This should be "Variable x is not defined when coming
                    #  from {bb}". But for this we need a way to associate BBs with
                    #  source locations.
                    raise GuppyError(
                        f"Variable `{x}` is not defined on all control-flow paths.",
                        use_bb.vars.used[x],
                    )
                raise GuppyError(f"Variable `{x}` is not defined", use_bb.vars.used[x])

    # Finally, we need to compute the signature of the basic block
    outputs = [
        [ctx.locals[x] for x in cfg.live_before[succ] if x in ctx.locals]
        for succ in bb.successors
    ]

    # Also prepare the successor list so we can fill it in later
    checked_bb = CheckedBB(
        bb.idx, checked_cfg, checked_stmts, sig=Signature(inputs, outputs)
    )
    checked_bb.successors = [None] * len(bb.successors)  # type: ignore[list-item]
    checked_bb.branch_pred = bb.branch_pred
    return checked_bb


def check_rows_match(row1: Row[Variable], row2: Row[Variable], bb: BB) -> None:
    """Checks that the types of two rows match up.

    Otherwise, an error is thrown, alerting the user that a variable has different
    types on different control-flow paths.
    """
    map1, map2 = {v.name: v for v in row1}, {v.name: v for v in row2}
    assert map1.keys() == map2.keys()
    for x in map1:
        v1, v2 = map1[x], map2[x]
        if v1.ty != v2.ty:
            # In the error message, we want to mention the variable that was first
            # defined at the start.
            if (
                v1.defined_at
                and v2.defined_at
                and line_col(v2.defined_at) < line_col(v1.defined_at)
            ):
                v1, v2 = v2, v1
            # We shouldn't mention temporary variables (starting with `%`)
            # in error messages:
            ident = "Expression" if v1.name.startswith("%") else f"Variable `{v1.name}`"
            raise GuppyError(
                f"{ident} can refer to different types: "
                f"`{v1.ty}` (at {{}}) vs `{v2.ty}` (at {{}})",
                bb.containing_cfg.live_before[bb][v1.name].vars.used[v1.name],
                [v1.defined_at, v2.defined_at],
            )


T = TypeVar("T")


def reverse_enumerate(xs: list[T]) -> Iterator[tuple[int, T]]:
    """Enumerates a list in reverse order.

    Equivalent to `reversed(list(enumerate(data)))` without creating an intermediate
    list.
    """
    for i in range(len(xs) - 1, -1, -1):
        yield i, xs[i]



import ast
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from hugr.serialization import ops

from guppylang.ast_util import AstNode, get_type, with_loc, with_type
from guppylang.checker.core import Context, Globals
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompiledGlobals, DFContainer
from guppylang.definition.common import ParsableDef
from guppylang.definition.value import CompiledCallableDef
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.hugr_builder.hugr import Hugr, OutPortV
from guppylang.nodes import GlobalCall
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import FunctionType, NoneType, Type, type_to_row


@dataclass(frozen=True)
class RawCustomFunctionDef(ParsableDef):
    """A raw custom function definition provided by the user.

    Custom functions provide their own checking and compilation logic using a
    `CustomCallChecker` and a `CustomCallCompiler`.

    The raw definition stores exactly what the user has written (i.e. the AST together
    with the provided checker and compiler), without inspecting the signature.
    """

    defined_at: ast.FunctionDef
    call_checker: "CustomCallChecker"
    call_compiler: "CustomCallCompiler"

    # Whether the function may be used as a higher-order value. This is only possible
    # if a static type for the function is provided.
    higher_order_value: bool

    description: str = field(default="function", init=False)

    def parse(self, globals: "Globals") -> "CustomFunctionDef":
        """Parses and checks the user-provided signature of the custom function.

        The signature is optional if custom type checking logic is provided by the user.
        However, note that the signature annotation is required, if we want to use the
        function as a higher-order value.

        If no signature is provided, we fill in the dummy signature `() -> ()`. This
        type will never be inspected, since we rely on the provided custom checking
        code. The only information we need to access is that it's a function type and
        that there are no unsolved existential vars.
        """
        # Type annotations are needed if we rely on the default call checker or want
        # to allow the usage of the function as a higher-order value
        requires_type_annotation = (
            isinstance(self.call_checker, DefaultCallChecker) or self.higher_order_value
        )
        has_type_annotation = self.defined_at.returns or any(
            arg.annotation for arg in self.defined_at.args.args
        )

        if requires_type_annotation and not has_type_annotation:
            raise GuppyError(
                f"Type signature for function `{self.name}` is required. "
                "Alternatively, try passing `higher_order_value=False` on definition.",
                self.defined_at,
            )

        ty = (
            check_signature(self.defined_at, globals)
            if requires_type_annotation
            else FunctionType([], NoneType())
        )
        return CustomFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            ty,
            self.call_checker,
            self.call_compiler,
            self.higher_order_value,
        )

    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> list[OutPortV]:
        """Compiles a call to the function."""
        self.call_compiler._setup(type_args, dfg, graph, globals, node)
        return self.call_compiler.compile(args)


@dataclass(frozen=True)
class CustomFunctionDef(CompiledCallableDef):
    """A custom function with parsed and checked signature."""

    defined_at: AstNode
    call_checker: "CustomCallChecker"
    call_compiler: "CustomCallCompiler"
    ty: FunctionType
    higher_order_value: bool

    description: str = field(default="function", init=False)

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, subst = self.call_checker.check(args, ty)
        return with_type(ty, with_loc(node, new_node)), subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, ty = self.call_checker.synthesize(args)
        return with_type(ty, with_loc(node, new_node)), ty

    def load_with_args(
        self,
        type_args: Inst,
        dfg: "DFContainer",
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> OutPortV:
        """Loads the custom function as a value into a local dataflow graph.

        This will place a `FunctionDef` node into the Hugr module and loads it into the
        DFG. This operation will fail the function is not allowed to be used as a
        higher-order value.
        """
        # TODO: This should be raised during checking, not compilation!
        if not self.higher_order_value:
            raise GuppyError(
                "This function does not support usage in a higher-order context",
                node,
            )
        assert len(self.ty.params) == len(type_args)

        # We create a `FunctionDef` that takes some inputs, compiles a call to the
        # function, and returns the results. If the function signature is polymorphic,
        # we explicitly monomorphise here and invoke the call compiler with the
        # inferred type args.
        fun_ty = self.ty.instantiate(type_args)
        def_node = graph.add_def(fun_ty, dfg.node, self.name)
        with graph.parent(def_node):
            _, inp_ports = graph.add_input_with_ports(list(fun_ty.inputs))
            returns = self.compile_call(
                inp_ports,
                type_args,
                DFContainer(graph, def_node),
                graph,
                globals,
                node,
            )
            graph.add_output(returns)

        # Finally, load the function into the local DFG. We already monomorphised, so we
        # can load with empty type args
        return graph.add_load_function(def_node.out_port(0), [], dfg.node).out_port(0)

    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> list[OutPortV]:
        """Compiles a call to the function."""
        self.call_compiler._setup(type_args, dfg, graph, globals, node)
        return self.call_compiler.compile(args)


class CustomCallChecker(ABC):
    """Abstract base class for custom function call type checkers."""

    ctx: Context
    node: AstNode
    func: CustomFunctionDef

    def _setup(self, ctx: Context, node: AstNode, func: CustomFunctionDef) -> None:
        self.ctx = ctx
        self.node = node
        self.func = func

    @abstractmethod
    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        """Checks the return value against a given type.

        Returns a (possibly) transformed and annotated AST node for the call.
        """

    @abstractmethod
    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        """Synthesizes a type for the return value of a call.

        Also returns a (possibly) transformed and annotated argument list.
        """


class CustomCallCompiler(ABC):
    """Abstract base class for custom function call compilers."""

    type_args: Inst
    dfg: DFContainer
    graph: Hugr
    globals: CompiledGlobals
    node: AstNode

    def _setup(
        self,
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> None:
        self.type_args = type_args
        self.dfg = dfg
        self.graph = graph
        self.globals = globals
        self.node = node

    @abstractmethod
    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        """Compiles a custom function call and returns the resulting ports."""


class DefaultCallChecker(CustomCallChecker):
    """Checks function calls by comparing to a type signature."""

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.func.ty, args, ty, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), subst

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.func.ty, args, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), ty


class NotImplementedCallCompiler(CustomCallCompiler):
    """Call compiler for custom functions that are already lowered during checking.

    For example, the custom checker could replace the call with a series of calls to
    other functions. In that case, the original function will no longer be present and
    thus doesn't need to be compiled.
    """

    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        raise InternalGuppyError("Function should have been removed during checking")


class OpCompiler(CustomCallCompiler):
    """Call compiler for functions that are directly implemented via Hugr ops."""

    op: ops.OpType

    def __init__(self, op: ops.OpType) -> None:
        self.op = op

    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        node = self.graph.add_node(
            self.op.model_copy(deep=True), inputs=args, parent=self.dfg.node
        )
        return_ty = get_type(self.node)
        return [node.add_out_port(ty) for ty in type_to_row(return_ty)]


class NoopCompiler(CustomCallCompiler):
    """Call compiler for functions that are noops."""

    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        return args



import ast
from collections.abc import Sequence

from guppylang.ast_util import (
    AstNode,
    set_location_from,
    shift_loc,
)
from guppylang.checker.core import Globals
from guppylang.definition.parameter import ParamDef
from guppylang.definition.ty import TypeDef
from guppylang.error import GuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.const import ConstValue
from guppylang.tys.param import Parameter, TypeParam
from guppylang.tys.ty import NoneType, NumericType, TupleType, Type


def arg_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter] | None = None,
) -> Argument:
    """Turns an AST expression into an argument."""
    # A single identifier
    if isinstance(node, ast.Name):
        x = node.id
        if x not in globals:
            raise GuppyError("Unknown identifier", node)
        match globals[x]:
            # Either a defined type (e.g. `int`, `bool`, ...)
            case TypeDef() as defn:
                return TypeArg(defn.check_instantiate([], globals, node))
            # Or a parameter (e.g. `T`, `n`, ...)
            case ParamDef() as defn:
                if param_var_mapping is None:
                    raise GuppyError(
                        "Free type variable. Only function types can be generic", node
                    )
                if x not in param_var_mapping:
                    param_var_mapping[x] = defn.to_param(len(param_var_mapping))
                return param_var_mapping[x].to_bound()
            case defn:
                raise GuppyError(
                    f"Expected a type, got {defn.description} `{defn.name}`", node
                )

    # A parametrised type, e.g. `list[??]`
    if isinstance(node, ast.Subscript) and isinstance(node.value, ast.Name):
        x = node.value.id
        if x in globals:
            defn = globals[x]
            if isinstance(defn, TypeDef):
                arg_nodes = (
                    node.slice.elts
                    if isinstance(node.slice, ast.Tuple)
                    else [node.slice]
                )
                # Hack: Flatten argument lists to support the `Callable` type. For
                # example, we turn `Callable[[int, int], bool]` into
                # `Callable[int, int, bool]`.
                # TODO: We can get rid of this once we added support for variadic params
                arg_nodes = [
                    n
                    for arg in arg_nodes
                    for n in (arg.elts if isinstance(arg, ast.List) else (arg,))
                ]
                args = [
                    arg_from_ast(arg_node, globals, param_var_mapping)
                    for arg_node in arg_nodes
                ]
                ty = defn.check_instantiate(args, globals, node)
                return TypeArg(ty)
            # We don't allow parametrised variables like `T[int]`
            if isinstance(defn, ParamDef):
                raise GuppyError(
                    f"Variable `{x}` is not parameterized. Higher-kinded types are not "
                    f"supported",
                    node,
                )

    # We allow tuple types to be written as `(int, bool)`
    if isinstance(node, ast.Tuple):
        ty = TupleType(
            [type_from_ast(el, globals, param_var_mapping) for el in node.elts]
        )
        return TypeArg(ty)

    # `None` is represented as a `ast.Constant` node with value `None`
    if isinstance(node, ast.Constant) and node.value is None:
        return TypeArg(NoneType())

    # Integer literals are turned into nat args since these are the only ones we support
    # right now.
    # TODO: Once we also have int args etc, we need proper inference logic here
    if isinstance(node, ast.Constant) and isinstance(node.value, int):
        # Fun fact: int ast.Constant values are never negative since e.g. `-5` is a
        # `ast.UnaryOp` negation of a `ast.Constant(5)`
        assert node.value >= 0
        nat_ty = NumericType(NumericType.Kind.Nat)
        return ConstArg(ConstValue(nat_ty, node.value))

    # Finally, we also support delayed annotations in strings
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        try:
            [stmt] = ast.parse(node.value).body
            if not isinstance(stmt, ast.Expr):
                raise GuppyError("Invalid Guppy type", node)
            set_location_from(stmt, loc=node)
            shift_loc(
                stmt,
                delta_lineno=node.lineno - 1,  # -1 since lines start at 1
                delta_col_offset=node.col_offset + 1,  # +1 to remove the `"`
            )
            return arg_from_ast(stmt.value, globals, param_var_mapping)
        except (SyntaxError, ValueError):
            raise GuppyError("Invalid Guppy type", node) from None

    raise GuppyError("Not a valid type argument", node)


_type_param = TypeParam(0, "T", True)


def type_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter] | None = None,
) -> Type:
    """Turns an AST expression into a Guppy type."""
    # Parse an argument and check that it's valid for a `TypeParam`
    arg = arg_from_ast(node, globals, param_var_mapping)
    return _type_param.check_arg(arg, node).ty


def type_row_from_ast(node: ast.expr, globals: "Globals") -> Sequence[Type]:
    """Turns an AST expression into a Guppy type row.

    This is needed to interpret the return type annotation of functions.
    """
    # The return type `-> None` is represented in the ast as `ast.Constant(value=None)`
    if isinstance(node, ast.Constant) and node.value is None:
        return []
    ty = type_from_ast(node, globals)
    if isinstance(ty, TupleType):
        return ty.element_types
    else:
        return [ty]



import functools
from typing import TYPE_CHECKING

from guppylang.checker.cfg_checker import CheckedBB, CheckedCFG, Row, Signature
from guppylang.checker.core import Place, Variable
from guppylang.compiler.core import (
    CompiledGlobals,
    DFContainer,
    is_return_var,
    return_var,
)
from guppylang.compiler.expr_compiler import ExprCompiler
from guppylang.compiler.stmt_compiler import StmtCompiler
from guppylang.hugr_builder.hugr import CFNode, Hugr, Node, OutPortV
from guppylang.tys.builtin import is_bool_type
from guppylang.tys.ty import SumType, row_to_type, type_to_row

if TYPE_CHECKING:
    from collections.abc import Sequence


def compile_cfg(
    cfg: CheckedCFG[Place], graph: Hugr, parent: Node, globals: CompiledGlobals
) -> None:
    """Compiles a CFG to Hugr."""
    insert_return_vars(cfg)

    blocks: dict[CheckedBB[Place], CFNode] = {}
    for bb in cfg.bbs:
        blocks[bb] = compile_bb(bb, graph, parent, bb == cfg.entry_bb, globals)
    for bb in cfg.bbs:
        for succ in bb.successors:
            graph.add_edge(blocks[bb].add_out_port(), blocks[succ].in_port(None))


def compile_bb(
    bb: CheckedBB[Place],
    graph: Hugr,
    parent: Node,
    is_entry: bool,
    globals: CompiledGlobals,
) -> CFNode:
    """Compiles a single basic block to Hugr."""
    inputs = bb.sig.input_row if is_entry else sort_vars(bb.sig.input_row)

    # The exit BB is completely empty
    if len(bb.successors) == 0:
        assert len(bb.statements) == 0
        return graph.add_exit([v.ty for v in inputs], parent)

    # Otherwise, we use a regular `Block` node
    block = graph.add_block(parent)

    # Add input node and compile the statements
    inp = graph.add_input(output_tys=[v.ty for v in inputs], parent=block)
    dfg = DFContainer(graph, block)
    for i, v in enumerate(inputs):
        dfg[v] = inp.out_port(i)
    dfg = StmtCompiler(graph, globals).compile_stmts(bb.statements, dfg)

    # If we branch, we also have to compile the branch predicate
    if len(bb.successors) > 1:
        assert bb.branch_pred is not None
        branch_port = ExprCompiler(graph, globals).compile(bb.branch_pred, dfg)
    else:
        # Even if we don't branch, we still have to add a `Sum(())` predicates
        branch_port = graph.add_tag(
            variants=[[]], tag=0, inputs=[], parent=block
        ).out_port(0)

    # Finally, we have to add the block output.
    outputs: Sequence[Place]
    if len(bb.successors) == 1:
        # The easy case is if we don't branch: We just output all variables that are
        # specified by the signature
        [outputs] = bb.sig.output_rows
    else:
        # If we branch and the branches use the same places, then we can use a
        # regular output
        first, *rest = bb.sig.output_rows
        if all({p.id for p in first} == {p.id for p in r} for r in rest):
            outputs = first
        else:
            # Otherwise, we have to output a TupleSum: We put all non-linear variables
            # into the branch TupleSum and all linear variables in the normal output
            # (since they are shared between all successors). This is in line with the
            # ordering on variables which puts linear variables at the end. The only
            # exception are return vars which must be outputted in order.
            branch_port = choose_vars_for_tuple_sum(
                graph=graph,
                unit_sum=branch_port,
                output_vars=[
                    [
                        v
                        for v in sort_vars(row)
                        if not v.ty.linear or is_return_var(str(v))
                    ]
                    for row in bb.sig.output_rows
                ],
                dfg=dfg,
            )
            outputs = [v for v in first if v.ty.linear and not is_return_var(str(v))]

    graph.add_output(
        inputs=[branch_port] + [dfg[v] for v in sort_vars(outputs)], parent=block
    )
    return block


def insert_return_vars(cfg: CheckedCFG[Place]) -> None:
    """Patches a CFG by annotating dummy return variables in the BB signatures.

    The statement compiler turns `return` statements into assignments of dummy variables
    `%ret0`, `%ret1`, etc. We update the exit BB signature to make sure they are
    correctly outputted.
    """
    return_vars = [
        Variable(return_var(i), ty, None)
        for i, ty in enumerate(type_to_row(cfg.output_ty))
    ]
    # Before patching, the exit BB shouldn't take any inputs
    assert len(cfg.exit_bb.sig.input_row) == 0
    cfg.exit_bb.sig = Signature(return_vars, cfg.exit_bb.sig.output_rows)
    # Also patch the predecessors
    for pred in cfg.exit_bb.predecessors:
        # The exit BB will be the only successor
        assert len(pred.sig.output_rows) == 1
        assert len(pred.sig.output_rows[0]) == 0
        pred.sig = Signature(pred.sig.input_row, [return_vars])


def choose_vars_for_tuple_sum(
    graph: Hugr, unit_sum: OutPortV, output_vars: list[Row[Place]], dfg: DFContainer
) -> OutPortV:
    """Selects an output based on a TupleSum.

    Given `unit_sum: Sum(*(), *(), ...)` and output variable rows `#s1, #s2, ...`,
    constructs a TupleSum value of type `Sum(#s1, #s2, ...)`.
    """
    assert isinstance(unit_sum.ty, SumType) or is_bool_type(unit_sum.ty)
    assert len(output_vars) == (
        len(unit_sum.ty.element_types) if isinstance(unit_sum.ty, SumType) else 2
    )
    assert all(not v.ty.linear for var_row in output_vars for v in var_row)
    conditional = graph.add_conditional(cond_input=unit_sum, inputs=[], parent=dfg.node)
    tys = [[v.ty for v in var_row] for var_row in output_vars]
    for i, var_row in enumerate(output_vars):
        case = graph.add_case(conditional)
        graph.add_input(output_tys=[], parent=case)
        inputs = [dfg[v] for v in var_row]
        tag = graph.add_tag(variants=tys, tag=i, inputs=inputs, parent=case).out_port(0)
        graph.add_output(inputs=[tag], parent=case)
    return conditional.add_out_port(SumType([row_to_type(row) for row in tys]))


def compare_var(p1: Place, p2: Place) -> int:
    """Defines a `<` order on variables.

    We use this to determine in which order variables are outputted from basic blocks.
    We need to output linear variables at the end, so we do a lexicographic ordering of
    linearity and name. The only exception are return vars which must be outputted in
    order.
    """
    x, y = str(p1), str(p2)
    if is_return_var(x) and is_return_var(y):
        return -1 if x < y else 1
    return -1 if (p1.ty.linear, x) < (p2.ty.linear, y) else 1


def sort_vars(row: Row[Place]) -> list[Place]:
    """Sorts a row of variables.

    This determines the order in which they are outputted from a BB.
    """
    return sorted(row, key=functools.cmp_to_key(compare_var))
